//**************************************************************************/
// DESCRIPTION: 
//	IKFKSolver 
//This solver does not uses the default system used by the other solvers because it needs
//to have full control of the nodes and their controllers. Instead, this solver uses a main
//class called IKFKSolver that has the duty of calculate the transform values of the nodes 
//that belong to the system, in order to use that some specific controllers had to be created
//in order to work as an interface to this solver class (FKCtrl and IKCtrl)
//as soon as the tm value is requested to the controllers of the node, they ask for it to the
//solver, the IKFKSolver "solves" the values and those are stored in a variable inside of the 
//controllers, the controllers use that variable to retrieve the value (that variable is called
//"value"). the solver solves the values only once unless the system or the time has changed.

//this solver works with only 3 "FK Nodes" (bones/joints), those are: main, sec and last nodes,
//the last node is used as "End Bone"
//It also have several "IK Nodes" those are: goal, target and goal parents.
//The FK nodes uses FKCtrl Controllers and the IK use IKCtrl controllers.

// AUTHOR: Felix Joleanes, www.joleanes.com
//***************************************************************************/

#include "ikfksolver.h"
#include "fkctrl.h"
#include "ikctrl.h"


static IKFKSolverClassDesc IKFKSolverDesc;
ClassDesc2* GetIKFKSolverDesc() { return &IKFKSolverDesc;}
static IKFKSolverClassAccessor IKFKSolverAccessor;

static ParamBlockDesc2 IKFKSolver_param_blk ( IKFKSolver_params, _T("params"),  0, &IKFKSolverDesc, 
	P_AUTO_CONSTRUCT + P_AUTO_UI +P_MULTIMAP, PBLOCK_REF,
	3,
	//rollout1
	IKFKSolver_params,IDD_PANEL_PARAMS, IDS_PARAMS, BEGIN_EDIT_MOTION, 0, NULL,
	//rollout2
	IKFKSolver_options,IDD_PANEL_OPTIONS, IDS_OPTIONS, BEGIN_EDIT_MOTION, 0, NULL,
	//rollout3
	IKFKSolver_display,IDD_PANEL_DISPLAY, IDS_DISPLAY, BEGIN_EDIT_MOTION, 0, NULL,

	// params
	
	pb_IKFKBlend, 		_T("IKFKBlend"), 		TYPE_PCNT_FRAC, 	P_ANIMATABLE+P_RESET_DEFAULT, 	IDS_IKFKBLEND, 
		p_default, 		0.0f, 
		p_range,		0.0f,		100.0f,
		p_ui, 			IKFKSolver_params,	TYPE_SPINNER,		EDITTYPE_FLOAT, IDC_EDIT_IKFKBLEND, IDC_SPIN_IKFKBLEND, 1.0f,
		end,
	pb_IKFKSwivelAngle, _T("SwivelAngle"), 		TYPE_ANGLE, 	P_ANIMATABLE+P_RESET_DEFAULT, 	IDS_IKFKSWIVEL, 
		p_default, 		0.0f, 
		p_range,		-999999.999999f,		999999.999999f,
		p_ui, 			IKFKSolver_params,	TYPE_SPINNER,		EDITTYPE_FLOAT, IDC_EDIT_IKFKSWIVEL, IDC_SPIN_IKFKSWIVEL, 1.0f,
		end,
	pb_scaleMult,		_T("ScaleMult"), 		TYPE_PCNT_FRAC, 	P_ANIMATABLE+P_RESET_DEFAULT, 	IDS_SCALEMULT, 
		p_default, 		1.0f, 
		p_range,		0.0f,		999999.999999f,
		p_ui, 			IKFKSolver_params,	TYPE_SPINNER,		EDITTYPE_FLOAT, IDC_EDIT_SCALEMULT, IDC_SPIN_SCALEMULT, 1.0f,
		end,
	pb_IKStretch,		_T("IKStretch"), 		TYPE_PCNT_FRAC, 	P_ANIMATABLE+P_RESET_DEFAULT, 	IDS_IKSTRETCH, 
		p_default, 		0.0f, 
		p_range,		0.0f,		100.0f,
		p_ui, 			IKFKSolver_params,	TYPE_SPINNER,		EDITTYPE_FLOAT, IDC_EDIT_IKSTRETCH, IDC_SPIN_IKSTRETCH, 1.0f,
		end,
	pb_matchMainPreferred,	_T(""),	TYPE_BOOL,	P_RESET_DEFAULT,	IDS_MATCHMAINPREFERRED,
		p_default,	FALSE,
		p_ui,		IKFKSolver_options,	TYPE_SINGLECHEKBOX,		IDC_CHECK_MATCHMAINPREFERRED,
		end,
	pb_matchSecPreferred,	_T(""),	TYPE_BOOL,	P_RESET_DEFAULT,	IDS_MATCHSECPREFERRED,
		p_default,	FALSE,
		p_ui,		IKFKSolver_options,	TYPE_SINGLECHEKBOX,		IDC_CHECK_MATCHSECPREFERRED,
		end,
	pb_matchScaleMult,	_T(""),	TYPE_BOOL,	P_RESET_DEFAULT,	IDS_MATCHSCALEMULT,
		p_default,	TRUE,
		p_ui,		IKFKSolver_options,	TYPE_SINGLECHEKBOX,		IDC_CHECK_MATCHSCALEMULT,
		end,
	pb_parentSpace, 	_T(""),		TYPE_INT,		P_RESET_DEFAULT,				IDS_PARENTSPACE,
		p_default, 		1, 
		p_range, 		0, 1, 
		p_ui, 			IKFKSolver_options,		TYPE_RADIO, 2,  IDC_RADIO_PARENT_IK,IDC_RADIO_PARENT_START,  
		end,
	pb_prefTreshold,	_T(""), 		TYPE_ANGLE, 	P_RESET_DEFAULT, 	IDS_PREFTRESHOLD, 
		p_default, 		0.174533f, 
		p_range,		1.0f,		180.0f,
		end,
	pb_displayGoal,	_T(""),	TYPE_INT,	P_RESET_DEFAULT,	IDS_DISPLAYGOAL,
		p_default,		1,
		p_range, 		0, 2,
		p_ui, 			IKFKSolver_display,		TYPE_RADIO, 3,  IDC_RADIO_DISPLAY_GOAL0, IDC_RADIO_DISPLAY_GOAL1, IDC_RADIO_DISPLAY_GOAL2,
		end,
	pb_displayTarget,	_T(""),	TYPE_INT,	P_RESET_DEFAULT,	IDS_DISPLAYTARGET,
		p_default,		1,
		p_range, 		0, 2,
		p_ui, 			IKFKSolver_display,		TYPE_RADIO, 3,  IDC_RADIO_DISPLAY_TARGET0, IDC_RADIO_DISPLAY_TARGET1, IDC_RADIO_DISPLAY_TARGET2,
		end,
	pb_displayIKFK,	_T(""),	TYPE_BOOL,	P_RESET_DEFAULT,	IDS_DISPLAYIKFK,
		p_default,	FALSE,
		end,
	pb_nGoalParent, 	_T(""),		TYPE_INT,		P_RESET_DEFAULT,				IDS_NGOALPARENT,
		p_default, 		0,  
		p_range,		0, 29999,
		p_accessor,		&IKFKSolverAccessor,
		end,
	pb_showAllKeys,	_T(""),	TYPE_BOOL,	0,	IDS_SHOWALLKEYS,
		p_default,	TRUE,
		p_ui,		IKFKSolver_options,	TYPE_SINGLECHEKBOX,		IDC_CHECK_SHOWALLKEYS,		
		end,
	pb_matchIKFK,	_T(""),	TYPE_BOOL,	P_RESET_DEFAULT,	IDS_MATCHIKFK,
		p_default,	TRUE,
		p_ui,		IKFKSolver_options,	TYPE_SINGLECHEKBOX,		IDC_CHECK_MATCHIKFK,
		end,
	pb_matchTargetSwivel,	_T(""),	TYPE_BOOL,	P_RESET_DEFAULT,	IDS_MATCHTARGETSWIVEL,
		p_default,	TRUE,
		p_ui,		IKFKSolver_options,	TYPE_SINGLECHEKBOX,		IDC_CHECK_MATCHTARGETSWIVEL,
		end,
	end
	
	);
static INT_PTR CALLBACK AboutDlgProc(
		HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
switch (msg) {
	case WM_INITDIALOG:
		CenterWindow(hWnd, GetParent(hWnd));
		break;

	case WM_COMMAND:
		switch (LOWORD(wParam)) {
			case IDOK:
				EndDialog(hWnd,1);
				break;
			}
		break;
	case WM_CLOSE:
			EndDialog(hWnd,1);		
		break;
	
	default:
		return FALSE;
	}
return TRUE;
}

void SetTMController(INode *theNode,Control *theControl)
{
	if (theNode)
	{
		//if the node is selected it has to be unselected and selected again to refresh ui in command panel
		BOOL isSelected=theNode->Selected();
		if(isSelected)
			GetCOREInterface()->DeSelectNode(theNode);
		Control *prevCtrl=theNode->GetTMController();
		
		//if the controller was part of the solver make sure it's removed from the scene
		if (prevCtrl->ClassID()==FKCtrl_CLASS_ID)
		{
			Control *FK=((FKCtrl*)prevCtrl)->FK;
			if (FK)
				FK->TransferReferences(prevCtrl,TRUE);
		}
		else if (prevCtrl->ClassID()==IKCtrl_CLASS_ID)
		{
			Control *IK=((IKCtrl*)prevCtrl)->IK;
			if (IK)
				IK->TransferReferences(prevCtrl,TRUE);
		}

		theNode->SetTMController(theControl);
		if (isSelected)
			GetCOREInterface()->SelectNode(theNode);
	}
}

INT_PTR IKFKSolverParamsDlgProc::DlgProc (TimeValue t, IParamMap2 *map, HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	if (solver)
	{
		switch(msg)
		{
		case WM_INITDIALOG:
			{
				INode *targetNode=solver->GetTargetNode();
				ICustEdit *iEdit = GetICustEdit(GetDlgItem(solver->hwnd1,IDC_EDIT_IKFKSWIVEL));
				iEdit->Enable(!targetNode);
				ReleaseICustEdit(iEdit);
				ISpinnerControl *iSpinner=GetISpinner(GetDlgItem(solver->hwnd1,IDC_SPIN_IKFKSWIVEL));
				iSpinner->Enable(!targetNode);
				ReleaseISpinner(iSpinner);
			}
			break;
		case WM_COMMAND:
				switch (LOWORD(wParam))
				{
				case IDC_BUTTON_FK:
					{
						BOOL isNotHolding=!theHold.Holding();
						if (isNotHolding)
							theHold.Begin();
						BOOL canUpdate=solver->CanUpdate();
						solver->Match(t);
						if (solver->pblock)
							solver->pblock->SetValue(pb_IKFKBlend,t,0.0f);
						if (isNotHolding)
							theHold.Accept(_T("IKFK set FK"));
					}				
					break;
				case IDC_BUTTON_IK:
					{
						BOOL isNotHolding=!theHold.Holding();
						if (isNotHolding)
							theHold.Begin();
						solver->Match(t);
						if (solver->pblock)
							solver->pblock->SetValue(pb_IKFKBlend,t,1.0f);
						if (isNotHolding)
							theHold.Accept(_T("IKFK set IK"));
					}
					break;
				}
			break;
		}
	}
	return FALSE;
}


INT_PTR IKFKSolverOptionsDlgProc::DlgProc (TimeValue t, IParamMap2 *map, HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	if (solver)
	{
		switch(msg)
		{
		case WM_INITDIALOG:
			{
				ICustButton *iBut = GetICustButton(GetDlgItem(solver->hwnd2,IDC_BUTTON_TARGET));
				if (iBut)
				{
					iBut->SetType(CBT_CHECK);
					iBut->SetHighlightColor(GREEN_WASH);
					INode *targetNode=solver->GetTargetNode();
					if (targetNode) 
						iBut->SetText(targetNode->GetName());
					else
						iBut->SetText(_T("None"));
					ReleaseICustButton(iBut);
				}
				iBut = GetICustButton(GetDlgItem(solver->hwnd2,IDC_BUTTON_GOALPARENT));
				if (iBut)
				{
					iBut->SetType(CBT_CHECK);
					iBut->SetHighlightColor(GREEN_WASH);
					INode *goalParentNode=solver->GetGoalParentNode(0);
					int nGoalParent=solver->GoalParentCount();
					if (nGoalParent && goalParentNode) 
						iBut->SetText(goalParentNode->GetName());
					else
						iBut->SetText(_T("None"));
					ReleaseICustButton(iBut);
				}
			}
			break;
		case WM_COMMAND:
				switch (LOWORD(wParam))
				{
				case IDC_BUTTON_TARGET:
					if (solver->ip)
						solver->ip->SetPickMode(new IKFKSolverPickNode(solver,PICK_TARGET));
					break;
				case IDC_BUTTON_GOALPARENT:
					if (solver->ip)
						solver->ip->SetPickMode(new IKFKSolverPickNode(solver,PICK_GOALPARENT));
					break;
				case IDC_BUTTON_REMOVE:
					{
						ICustButton *iBut = GetICustButton(GetDlgItem(solver->hwnd2,IDC_BUTTON_TARGET));
						if (iBut)
						{
							iBut->SetText(_T("None"));
							ReleaseICustButton(iBut);
						}
						ICustEdit *iEdit = GetICustEdit(GetDlgItem(solver->hwnd1,IDC_EDIT_IKFKSWIVEL));
						if (iEdit)
						{
							iEdit->Enable(TRUE);
							ReleaseICustEdit(iEdit);
						}
						ISpinnerControl *iSpinner=GetISpinner(GetDlgItem(solver->hwnd1,IDC_SPIN_IKFKSWIVEL));
						if (iSpinner)
						{
							iSpinner->Enable(TRUE);
							ReleaseISpinner(iSpinner);
						}
						solver->SetTargetNode(NULL);
					}
					break;
				case IDC_BUTTON_REMOVE_GOALPARENT:
					{
						TimeValue t=GetCOREInterface()->GetTime();
						ICustButton *iBut = GetICustButton(GetDlgItem(solver->hwnd2,IDC_BUTTON_GOALPARENT));
						if (iBut)
						{
							iBut->SetText(_T("None"));
							ReleaseICustButton(iBut);
						}
						solver->SetGoalParentNode(NULL);
					}
					break;
				case IDC_BUTTON_ABOUT:
					DialogBox(
							hInstance,
							MAKEINTRESOURCE(IDD_DIALOG),
							hWnd,
							AboutDlgProc);
					break;
				}
			break;
		}
	}
	return FALSE;
}
BOOL IKFKSolverPickNode::Filter(INode *node)
{
	if (solver)
		if (pickType==PICK_TARGET)
			return solver->isValidTarget(node);
		else
			return solver->isValidGoalParent(node);
	return FALSE;
}

BOOL IKFKSolverPickNode::HitTest(IObjParam *ip,HWND hWnd,ViewExp *vpt,IPoint2 m,int flags)
{	
	if (ip && ip->PickNode(hWnd,m,this))
		return TRUE;
	else
		return FALSE;
}

BOOL IKFKSolverPickNode::Pick(IObjParam *ip,ViewExp *vpt)
{
	INode *node=NULL;
	if (vpt)
		node= vpt->GetClosestHit();
	if (solver && node) 
	{
		ICustButton *iBut;
		if (pickType==PICK_TARGET)
		{
			solver->SetTargetNode(node);
			iBut = GetICustButton(GetDlgItem(solver->hwnd2,IDC_BUTTON_TARGET));
			if (iBut) 
			{
				iBut->SetText(node->GetName());
				ReleaseICustButton(iBut);
			}
			ICustEdit *iEdit = GetICustEdit(GetDlgItem(solver->hwnd1,IDC_EDIT_IKFKSWIVEL));
			if (iEdit) 
			{
				iEdit->Enable(FALSE);
				ReleaseICustEdit(iEdit);
			}
		}
		else
		{
			solver->SetGoalParentNode(node);
			iBut = GetICustButton(GetDlgItem(solver->hwnd2,IDC_BUTTON_GOALPARENT));
			if (iBut) 
			{
				iBut->SetText(node->GetName());
				ReleaseICustButton(iBut);
			}
		}
		return TRUE;
	}
	return FALSE;
}

void IKFKSolverPickNode::EnterMode(IObjParam *ip)
{
	if (solver)
	{
		ICustButton *iBut;
		if (pickType==PICK_TARGET)
		{
			iBut= GetICustButton(GetDlgItem(solver->hwnd2,IDC_BUTTON_TARGET));
			if (iBut)
			{
				iBut->SetCheck(TRUE);
				ReleaseICustButton(iBut);
			}
			GetCOREInterface()->PushPrompt(_T("Select Target Object"));
		}
		else
		{
			iBut= GetICustButton(GetDlgItem(solver->hwnd2,IDC_BUTTON_GOALPARENT));
			if (iBut) 
			{
				iBut->SetCheck(TRUE);
				ReleaseICustButton(iBut);
			}
			GetCOREInterface()->PushPrompt(_T("Select Target Object"));
		}
	}
}

void IKFKSolverPickNode::ExitMode(IObjParam *ip)
{
	if (solver)
	{
		if (pickType==PICK_TARGET)
		{
			ICustButton *iBut = GetICustButton(GetDlgItem(solver->hwnd2,IDC_BUTTON_TARGET));
			if (iBut) 
			{
				iBut->SetCheck(FALSE);
				ReleaseICustButton(iBut);
			}
			GetCOREInterface()->PopPrompt();
		}
		else
		{
			ICustButton *iBut = GetICustButton(GetDlgItem(solver->hwnd2,IDC_BUTTON_GOALPARENT));
			if (iBut)
			{
				iBut->SetCheck(FALSE);
				ReleaseICustButton(iBut);
			}
			GetCOREInterface()->PopPrompt();
		}
	}
}
int  IKFKSolverPostPatchProc::Proc (RemapDir &remap)
{
	if (solver)
	{
		IKFKSolver *clonedSolver=(IKFKSolver *)remap.FindMapping (solver);
		INode *goalNode=(INode *)remap.FindMapping (solver->GetGoalNode());
		INode *mainNode=(INode *)remap.FindMapping (solver->GetMainNode());
		INode *secNode=(INode *)remap.FindMapping (solver->GetSecNode());
		INode *lastNode=(INode *)remap.FindMapping (solver->GetLastNode());
		INode *targetNode=(INode *)remap.FindMapping (solver->GetTargetNode());

		int i,nGoalParent=solver->GoalParentCount();
		INode **goalParentNode=new INode*[nGoalParent];
		BOOL goalParentValid=TRUE;
		for (i=0;i<nGoalParent;i++)
		{
			goalParentNode[i]=(INode *)remap.FindMapping (solver->GetGoalParentNode(i));
			if (!goalParentNode[i])
				goalParentValid=FALSE;
		}
		

		//set the cloned references
		if (goalNode && mainNode && secNode && lastNode)
		{
			INodeTransformMonitor* nodeTransformMonitor;
			if (clonedSolver->goalNode)
			{
				nodeTransformMonitor = (INodeTransformMonitor*)(clonedSolver->goalNode->GetInterface(IID_NODETRANSFORMMONITOR));
				nodeTransformMonitor->SetNode(goalNode);
			}
			if (clonedSolver->mainNode)
			{
			nodeTransformMonitor = (INodeTransformMonitor*)(clonedSolver->mainNode->GetInterface(IID_NODETRANSFORMMONITOR));
			nodeTransformMonitor->SetNode(mainNode);
			}
			if (clonedSolver->secNode)
			{
			nodeTransformMonitor = (INodeTransformMonitor*)(clonedSolver->secNode->GetInterface(IID_NODETRANSFORMMONITOR));
			nodeTransformMonitor->SetNode(secNode);
			}
			if (clonedSolver->lastNode)
			{
			nodeTransformMonitor = (INodeTransformMonitor*)(clonedSolver->lastNode->GetInterface(IID_NODETRANSFORMMONITOR));
			nodeTransformMonitor->SetNode(lastNode);
			}
			if (clonedSolver->targetNode)
			{
			nodeTransformMonitor = (INodeTransformMonitor*)(clonedSolver->targetNode->GetInterface(IID_NODETRANSFORMMONITOR));
			nodeTransformMonitor->SetNode(targetNode);
			}

			//set the cloned goalparents only if all of them were cloned
			if (goalParentValid)
			{
				for (i=0;i<nGoalParent;i++)
				{
					clonedSolver->ReplaceReference(NSOLVER_REFS+i,(ReferenceTarget*)CreateInstance(REF_TARGET_CLASS_ID, NODETRANSFORMMONITOR_CLASS_ID));
					if (clonedSolver->goalParentNode[i])
					{
						nodeTransformMonitor = (INodeTransformMonitor*)(clonedSolver->goalParentNode[i]->GetInterface(IID_NODETRANSFORMMONITOR));
						nodeTransformMonitor->SetNode(goalParentNode[i]);
					}
				}
			}
			else
			{
				clonedSolver->SetGoalParentCount(0);
				for (i=0;i<nGoalParent;i++)
				{
					if (goalParentNode[i] && goalParentNode[i]->GetTMController()->ClassID()==IKCtrl_CLASS_ID)
					{
						Control *theCtrl=(Control*)((IKCtrl*)goalParentNode[i]->GetTMController())->IK;
						if (theCtrl)
							goalParentNode[i]->SetTMController(theCtrl);
						else
							goalParentNode[i]->SetTMController(NewDefaultMatrix3Controller());
					}
				}
			}
			clonedSolver->NotifyChange();
		}
		else //remove solver if it was not copied entirelly
		{
			Control *theCtrl=NULL;
			clonedSolver->SetCanUpdate(FALSE);
			clonedSolver->DeleteAllRefs();
			if (goalNode && goalNode->GetTMController()->ClassID()==IKCtrl_CLASS_ID)
			{
				theCtrl=(Control*)((IKCtrl*)goalNode->GetTMController())->IK;
				if (theCtrl)
					goalNode->SetTMController(theCtrl);
				else
					goalNode->SetTMController(NewDefaultMatrix3Controller());
			}
			if (mainNode && mainNode->GetTMController()->ClassID()==FKCtrl_CLASS_ID)
			{
				theCtrl=(Control*)((FKCtrl*)mainNode->GetTMController())->FK;
				if (theCtrl)
					mainNode->SetTMController(theCtrl);
				else
					mainNode->SetTMController(NewDefaultMatrix3Controller());
			}
			if (secNode && secNode->GetTMController()->ClassID()==FKCtrl_CLASS_ID)
			{
				theCtrl=(Control*)((FKCtrl*)secNode->GetTMController())->FK;
				if (theCtrl)
					secNode->SetTMController(theCtrl);
				else
					secNode->SetTMController(NewDefaultMatrix3Controller());
			}
			if (lastNode && lastNode->GetTMController()->ClassID()==FKCtrl_CLASS_ID)
			{
				theCtrl=(Control*)((FKCtrl*)lastNode->GetTMController())->FK;
				if (theCtrl)
					lastNode->SetTMController(theCtrl);
				else
					lastNode->SetTMController(NewDefaultMatrix3Controller());
			}
			if (targetNode && targetNode->GetTMController()->ClassID()==IKCtrl_CLASS_ID)
			{
				theCtrl=(Control*)((IKCtrl*)targetNode->GetTMController())->IK;
				if (theCtrl)
					targetNode->SetTMController(theCtrl);
				else
					targetNode->SetTMController(NewDefaultMatrix3Controller());
			}
			for (i=0;i<nGoalParent;i++)
			{
				if (goalParentNode[i] && goalParentNode[i]->GetTMController()->ClassID()==IKCtrl_CLASS_ID)
				{
					theCtrl=(Control*)((IKCtrl*)goalParentNode[i]->GetTMController())->IK;
					if (theCtrl)
						goalParentNode[i]->SetTMController(theCtrl);
					else
						goalParentNode[i]->SetTMController(NewDefaultMatrix3Controller());
				}
			}
			
		}
	}
	return 0;
}

void IKFKSolverClassAccessor::Set (PB2Value &v, ReferenceMaker *owner, ParamID id, int tabIndex, TimeValue t)
{
	if (id==pb_nGoalParent && owner)
	{
		IKFKSolver *solver=(IKFKSolver *)owner;
		if (solver)
		{
			ICustButton *iBut = GetICustButton(GetDlgItem(solver->hwnd2,IDC_BUTTON_GOALPARENT));
			if (iBut)
			{
				INode *goalParentNode=solver->GetGoalParentNode(0);
				if (v.i && goalParentNode)
					iBut->SetText(goalParentNode->GetName());
				else
					iBut->SetText(_T("None"));
				ReleaseICustButton(iBut);
			}
		}
	}
}

//NodeDeleteNotify (Global)
//This notify is used to remove the solver if any of its main nodes are deleted 
//(mainnode,secnode,lastnode,ikgoal)
void NodeDeleteNotify(void *param, NotifyInfo *info)
{
	IKFKSolver *theSolver=(IKFKSolver*)param;
	if (theSolver)
	{
		INode* goalNode=theSolver->GetGoalNode();
		INode* mainNode=theSolver->GetMainNode();
		INode* secNode=theSolver->GetSecNode();
		INode* lastNode=theSolver->GetLastNode();

		theSolver->SetTargetNode(NULL);
		theSolver->SetGoalParentNode(NULL);
		theSolver->SetCanUpdate(FALSE);

		if (goalNode)
			theSolver->SetGoalNode(NULL);
		if (mainNode)
			theSolver->SetMainNode(NULL);
		if (secNode)		
			theSolver->SetSecNode(NULL);
		if (lastNode)
			theSolver->SetLastNode(NULL);
	}
	UnRegisterNotification(NodeDeleteNotify,param, NOTIFY_SCENE_POST_DELETED_NODE);
}

//NodeMergeNotify (Global)
//This notify is used when merging to remove the solver if all the nodes were not merged
void NodeMergeNotify(void *param, NotifyInfo *info)
{
	IKFKSolver *theSolver=(IKFKSolver*)param;
	UnRegisterNotification(NodeMergeNotify,theSolver, NOTIFY_FILE_POST_MERGE);
	if (theSolver)
	{
		INode *mainNode=theSolver->GetMainNode();
		INode *secNode=theSolver->GetSecNode();
		INode *lastNode=theSolver->GetLastNode();
		INode *goalNode=theSolver->GetGoalNode();
		INode *targetNode=theSolver->GetTargetNode();
		Control *theCtrl;
		Control *localCtrl;
		if (!(mainNode && secNode && lastNode && goalNode))
		{
			for (int i=0;i<theSolver->GoalParentCount();i++)
			{
				INode *goalParentNode=theSolver->GetGoalParentNode(i);
				if (goalParentNode)
				{
					theCtrl=goalParentNode->GetTMController();
					if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
					{
						localCtrl=((IKCtrl*)theCtrl)->IK;
						if (localCtrl)
							goalParentNode->SetTMController(localCtrl);		
						else
							goalParentNode->SetTMController(NewDefaultMatrix3Controller());	
						goalParentNode->GetTMController()->NotifyDependents(FOREVER,PART_TM,REFMSG_CHANGE);
					}
				}
			}
			if (targetNode) 
			{
				theCtrl=targetNode->GetTMController();
				if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
				{
					localCtrl=((IKCtrl*)theCtrl)->IK;
					if (localCtrl)
						targetNode->SetTMController(localCtrl);		
					else
						targetNode->SetTMController(NewDefaultMatrix3Controller());
					targetNode->GetTMController()->NotifyDependents(FOREVER,PART_TM,REFMSG_CHANGE);
				}
			}
			if (mainNode) 
			{
				theCtrl=mainNode->GetTMController();
				if (theCtrl->ClassID()==FKCtrl_CLASS_ID)
				{
					localCtrl=((FKCtrl*)theCtrl)->FK;
					if (localCtrl)
						mainNode->SetTMController(localCtrl);
					else
						mainNode->SetTMController(NewDefaultMatrix3Controller());
					mainNode->GetTMController()->NotifyDependents(FOREVER,PART_TM,REFMSG_CHANGE);
				}
			}
			if (secNode) 
			{
				theCtrl=secNode->GetTMController();
				if (theCtrl->ClassID()==FKCtrl_CLASS_ID)
				{
					localCtrl=((FKCtrl*)theCtrl)->FK;
					if (localCtrl)
						secNode->SetTMController(((FKCtrl*)theCtrl)->FK);
					else
						secNode->SetTMController(NewDefaultMatrix3Controller());
					secNode->GetTMController()->NotifyDependents(FOREVER,PART_TM,REFMSG_CHANGE);
				}
			}
			if (lastNode)
			{
				theCtrl=lastNode->GetTMController();
				if (theCtrl->ClassID()==FKCtrl_CLASS_ID)
				{
					localCtrl=((FKCtrl*)theCtrl)->FK;
					if (localCtrl)
						lastNode->SetTMController(localCtrl);
					else
						lastNode->SetTMController(NewDefaultMatrix3Controller());
					lastNode->GetTMController()->NotifyDependents(FOREVER,PART_TM,REFMSG_CHANGE);
				}
			}
			if (goalNode) 
			{
				theCtrl=goalNode->GetTMController();
				if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
				{
					localCtrl=((IKCtrl*)theCtrl)->IK;
					if (localCtrl)
						goalNode->SetTMController(localCtrl);
					else
						goalNode->SetTMController(NewDefaultMatrix3Controller());
					goalNode->GetTMController()->NotifyDependents(FOREVER,PART_TM,REFMSG_CHANGE);
				}
			}
			
		}
		else
		{
			if (theSolver->GoalParentCount())
			{
				if (!theSolver->isValidGoalParent(theSolver->GetGoalParentNode(0)))
					theSolver->SetGoalParentNode(NULL);
			}
		}
	}
	
}
//SelectionNotify (Global)
//this notify is used to refresh the shared keys when selection changes
void SelectionNotify(void *param, NotifyInfo *info)
{
	if (param)
	{
		IKFKSolver *theSolver=(IKFKSolver*)param;
		BOOL isSelected=FALSE;
		INode *theNode;
		theNode=theSolver->GetMainNode();
		if (theNode && theNode->Selected())
			isSelected=TRUE;
		theNode=theSolver->GetSecNode();
		if (theNode && theNode->Selected())
			isSelected=TRUE;
		theNode=theSolver->GetLastNode();
		if (theNode && theNode->Selected())
			isSelected=TRUE;
		theNode=theSolver->GetGoalNode();
		if (theNode && theNode->Selected())
			isSelected=TRUE;
		theNode=theSolver->GetTargetNode();
		if (theNode && theNode->Selected())
			isSelected=TRUE;
		for (int i=0;i<theSolver->GoalParentCount();i++)
		{
			theNode=theSolver->GetGoalParentNode(i);
			if (theNode && theNode->Selected())
				isSelected=TRUE;
		}
		if (isSelected)
			theSolver->RefreshKeys();
	}
}

IKFKSolver::IKFKSolver(BOOL loading)
{
	pblock=NULL;
	goalNode=NULL;
	targetNode=NULL;
	mainNode=NULL;
	secNode=NULL;
	lastNode=NULL;
	SetCanUpdate(TRUE);
	canUpdateSwivel=TRUE;
	setTMCount=0;
	prevt=GetCOREInterface()->GetTime();
	maint=prevt;
	sect=prevt;
	goalt=prevt;
	hasChanged=TRUE;
	notProcessing=TRUE;
	mustRemoveGoalParent=FALSE;
	nGoalParentToRemove=0;
	allCtrls.Init();

	if (!loading)
	{
		BOOL isHolding=theHold.Holding() && !theHold.IsSuspended();
		if (isHolding) theHold.Suspend();
		IKFKSolverDesc.MakeAutoParamBlocks(this);
		ReplaceReference(GOAL_REF,(ReferenceTarget*)CreateInstance(REF_TARGET_CLASS_ID, NODETRANSFORMMONITOR_CLASS_ID));
		ReplaceReference(TARGET_REF,(ReferenceTarget*)CreateInstance(REF_TARGET_CLASS_ID, NODETRANSFORMMONITOR_CLASS_ID));
		ReplaceReference(MAIN_REF,(ReferenceTarget*)CreateInstance(REF_TARGET_CLASS_ID, NODETRANSFORMMONITOR_CLASS_ID));
		ReplaceReference(SEC_REF,(ReferenceTarget*)CreateInstance(REF_TARGET_CLASS_ID, NODETRANSFORMMONITOR_CLASS_ID));
		ReplaceReference(LAST_REF,(ReferenceTarget*)CreateInstance(REF_TARGET_CLASS_ID, NODETRANSFORMMONITOR_CLASS_ID));
		if (isHolding) theHold.Resume();
	}

	RegisterNotification(SelectionNotify,this, NOTIFY_SELECTIONSET_CHANGED);
}
/*
FUNCTION	IKFKSolver::IKFKSolver
Job:
	creates the ikfk system using the first and lastnode of the desired chain
Parameters:
	the selected first and last node of the ikfk chain
*/
IKFKSolver::IKFKSolver(INode *mainBone,INode *secBone,INode *lastBone,INode *goal)
{
	Interface *ip=GetCOREInterface();
	pblock=NULL;
	goalNode=NULL;
	targetNode=NULL;
	mainNode=NULL;
	secNode=NULL;
	lastNode=NULL;
	SetCanUpdate(FALSE);
	canUpdateSwivel=TRUE;
	notNotifyingKeySel=TRUE;
	setTMCount=0;
	prevt=ip->GetTime();
	maint=prevt;
	sect=prevt;
	goalt=prevt;
	notProcessing=TRUE;
	mustRemoveGoalParent=FALSE;
	nGoalParentToRemove=0;
	allCtrls.Init();

	BOOL isHolding=theHold.Holding() && !theHold.IsSuspended();
	if (isHolding) theHold.Suspend();
	IKFKSolverDesc.MakeAutoParamBlocks(this);
	ReplaceReference(GOAL_REF,(ReferenceTarget*)CreateInstance(REF_TARGET_CLASS_ID, NODETRANSFORMMONITOR_CLASS_ID));
	ReplaceReference(TARGET_REF,(ReferenceTarget*)CreateInstance(REF_TARGET_CLASS_ID, NODETRANSFORMMONITOR_CLASS_ID));
	ReplaceReference(MAIN_REF,(ReferenceTarget*)CreateInstance(REF_TARGET_CLASS_ID, NODETRANSFORMMONITOR_CLASS_ID));
	ReplaceReference(SEC_REF,(ReferenceTarget*)CreateInstance(REF_TARGET_CLASS_ID, NODETRANSFORMMONITOR_CLASS_ID));
	ReplaceReference(LAST_REF,(ReferenceTarget*)CreateInstance(REF_TARGET_CLASS_ID, NODETRANSFORMMONITOR_CLASS_ID));
	if (isHolding) theHold.Resume();

	isHolding=theHold.Holding();
	if (!isHolding)
		theHold.Begin();
	//Start matching, turn off animate first
	SuspendAnimate();
	AnimateOff();


	//Create the Goal if was not defined as parameter
	if (!goal)
	{
		Object *obj = (Object*)ip->CreateInstance(HELPER_CLASS_ID, Class_ID(POINTHELP_CLASS_ID,0));
		goal=ip->CreateObjectNode(obj);
		if (goal)
		{
			goal->SetWireColor(GOAL_COLOR);	
			MSTR newGoalName=MSTR(GetString(IDS_GOAL_NAME));
			ip->MakeNameUnique(newGoalName);
			goal->SetName(newGoalName);
		}
		else return;
	}

	SetGoalParentCount(0);
	SetGoalNode(goal);
	SetMainNode(mainBone);
	SetSecNode(secBone);
	SetLastNode(lastBone);

	//some matches
	Solve(prevt,FORCE_SOLVE+FULL_FKMAIN+FULL_FKSEC);
	MainPrefMatch(prevt);
	SecPrefMatch(prevt,TRUE);
	SwivelMatch(prevt);
	IKMatch(prevt);

	ip->SelectNode(goal);

	SetCanUpdate(TRUE);
	ResumeAnimate();


	RegisterNotification(SelectionNotify,this, NOTIFY_SELECTIONSET_CHANGED);

	if (!isHolding)
		theHold.Accept(_T("Create IKFKSolver"));
}
IKFKSolver::~IKFKSolver()
{
	UnRegisterNotification(NodeDeleteNotify,this, NOTIFY_SCENE_POST_DELETED_NODE);
	UnRegisterNotification(NodeMergeNotify,this, NOTIFY_FILE_POST_MERGE);
	UnRegisterNotification(SelectionNotify,this, NOTIFY_SELECTIONSET_CHANGED);
	DeleteAllRefs();
	pblock=NULL;
	goalNode=NULL;
	targetNode=NULL;
	mainNode=NULL;
	secNode=NULL;
	lastNode=NULL;
}
//to display controller params on the motion panel 
void IKFKSolver::BeginEditParams( IObjParam *ip, ULONG flags,Animatable *prev)
{
	if (pblock && (flags & BEGIN_EDIT_MOTION))
	{
		this->ip=ip;
		IKFKSolverDesc.BeginEditParams(ip,this,flags,prev);
		hwnd1=pblock->GetMap(IKFKSolver_params)->GetHWnd();
		hwnd2=pblock->GetMap(IKFKSolver_options)->GetHWnd();
		IKFKSolver_param_blk.SetUserDlgProc(IKFKSolver_params,new IKFKSolverParamsDlgProc(this));
		IKFKSolver_param_blk.SetUserDlgProc(IKFKSolver_options,new IKFKSolverOptionsDlgProc(this));
	}
}
// to undisplay 
void IKFKSolver::EndEditParams( IObjParam *ip, ULONG flags,Animatable *next )
{
	if (flags & END_EDIT_REMOVEUI)
	{
		IKFKSolverDesc.EndEditParams(ip,this,flags,next);
	}
}
RefTargetHandle IKFKSolver::Clone( RemapDir &remap )
{
	IKFKSolver *p=new IKFKSolver(FALSE);
	p->ReplaceReference(PBLOCK_REF,remap.CloneRef(pblock));
	BaseClone(this, p, remap);
	//set the node references after all the clone process has done
	//if the main, sec, last and goal nodes werent copied as well the solver should be removed
	//if (!remap.FindMapping (this))
		remap.AddPostPatchProc(new IKFKSolverPostPatchProc(this),TRUE);
	return p;
}


#if IS_SUPMAX2015
RefResult IKFKSolver::NotifyRefChanged(const Interval& changeInt, RefTargetHandle hTarget, 
			PartID& partID,  RefMessage message, BOOL propagate)
#else
RefResult IKFKSolver::NotifyRefChanged(Interval changeInt, RefTargetHandle hTarget, 
			PartID& partID,  RefMessage message)
#endif
{
	switch(message)
	{
	case REFMSG_CHANGE:
		{
			if (CanUpdate())
			{
				if (hTarget==pblock && pblock)
				{
					ParamID changing_param = pblock->LastNotifyParamID();
					IKFKSolver_param_blk.InvalidateUI(changing_param);

					if (changing_param==pb_showAllKeys)
					{
						RefreshKeys();
					}
					if (changing_param==pb_IKFKSwivelAngle && canUpdateSwivel) //if the Swivel changes the solver must do a ikmatch and a fkikmatch
					{
						/*Control* theCtrl;
						hasChanged=TRUE;
						SetCanUpdate(FALSE);
#if IS_SUPMAX2012
						theCtrl=pblock->GetControllerByIndex(pb_IKFKSwivelAngle);
#else
						theCtrl=pblock->GetController(pb_IKFKSwivelAngle);
#endif
						TimeValue t=GetCOREInterface()->GetTime();
						if (!theCtrl || theCtrl->NumKeys()==0 || theCtrl->GetKeyIndex(t)>=0)
						{
							Solve(t,FORCE_SOLVE+FULL_IKGOAL);
							FKMatch(t);
						}
						SetCanUpdate(TRUE);*/
					}
					else
						canUpdateSwivel=TRUE;
					NotifyChange();
				}
				if (hTarget==mainNode && partID==PART_TM)
				{
					hasChanged=TRUE;
					SetCanUpdate(FALSE);
					INode *targetNode=GetTargetNode();
					INode *goalNode=GetGoalNode();
					INode *goalParentNode=GetGoalParentNode(0);
					if (GoalParentCount() && goalParentNode)
						goalParentNode->GetTMController()->NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
					else if (goalNode)
						goalNode->GetTMController()->NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
					if (targetNode)
						targetNode->GetTMController()->NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
					SetCanUpdate(TRUE);
				}
				if (((hTarget==goalNode) || hTarget==targetNode) && partID==PART_TM)
				{
					SetCanUpdate(FALSE);
					INode *mainNode=GetMainNode();
					INode *secNode=GetSecNode();
					if (mainNode && secNode)
					{
						mainNode->GetTMController()->NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
						secNode->GetTMController()->NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
					}
					SetCanUpdate(TRUE);
					hasChanged=TRUE;
				}
			}
		}
		break;
	case REFMSG_NODE_CHANGED:
		{
			RefreshKeys();
			ICustButton *iBut = GetICustButton(GetDlgItem(hwnd2,IDC_BUTTON_TARGET));
			ICustEdit *iEdit = GetICustEdit(GetDlgItem(hwnd1,IDC_EDIT_IKFKSWIVEL));
			ISpinnerControl *iSpinner=GetISpinner(GetDlgItem(hwnd1,IDC_SPIN_IKFKSWIVEL));
			if (iBut && iEdit && iSpinner)
			{
				INode *targetNode=GetTargetNode();
				if (targetNode) 
					iBut->SetText(targetNode->GetName());
				else
					iBut->SetText(_T("None"));
				iEdit->Enable(!targetNode);
				iSpinner->Enable(!targetNode);
				ReleaseISpinner(iSpinner);
				ReleaseICustButton(iBut);
				ReleaseICustEdit(iEdit);
			}
			iBut = GetICustButton(GetDlgItem(hwnd2,IDC_BUTTON_GOALPARENT));
			if (iBut)
			{
				INode *goalParentNode=GetGoalParentNode(0);
				if (GoalParentCount() && goalParentNode) 
					iBut->SetText(goalParentNode->GetName());
				else
					iBut->SetText(_T("None"));
				ReleaseICustButton(iBut);
			}
		}
		break;
	case REFMSG_NODE_DELETED:
		{
			if (hTarget==mainNode || hTarget==secNode || hTarget==lastNode || hTarget==goalNode)
			{
				RegisterNotification(NodeDeleteNotify,this, NOTIFY_SCENE_POST_DELETED_NODE);
				SetCanUpdate(FALSE);
			}
		}
		break;
	case REFMSG_NODE_HANDLE_CHANGED:
		{
			//this notify is used to check if all the nodes were merged, if not then delete the solver
			//the solver deletion can't be done here because it will cause max to crash
			RegisterNotification(NodeMergeNotify,this, NOTIFY_FILE_POST_MERGE);
		}
		break;
	case REFMSG_KEY_SELECTION_CHANGED:
		{
			if (hTarget==pblock && pblock)
				NotifyKeySelChange();
		}
		break;
	}
	return REF_STOP;
}
RefTargetHandle IKFKSolver::GetReference(int i)
{
	switch(i)
	{
	case PBLOCK_REF:
		return pblock;
		break;
	case GOAL_REF:
		return goalNode;
		break;
	case TARGET_REF:
		return targetNode;
		break;
	case MAIN_REF:
		return mainNode;
		break;
	case SEC_REF:
		return secNode;
		break;
	case LAST_REF:
		return lastNode;
		break;
	case NONUSED_REF:
		return NULL;
		break;
	default:
		if (i>0 && i<(NSOLVER_REFS+GoalParentCount()))
			return goalParentNode[i-NSOLVER_REFS];
		else
			return NULL;
		break;
	}
}
void IKFKSolver::SetReference(int i, RefTargetHandle rtarg)
{
	switch(i)
	{
	case PBLOCK_REF:
		{
			pblock=(IParamBlock2*)rtarg;
			if (pblock) //initialize goal parent node containers
			{
				int n;
				pblock->GetValue(pb_nGoalParent,0,n,FOREVER);
				goalParentNode.SetCount(n);
				for (int i=0; i<n; i++)
					goalParentNode[i]=NULL;
			}
		}
		//return;
		break;
	case GOAL_REF:
		goalNode=(ReferenceTarget*)rtarg;
		break;
	case TARGET_REF:
		{
			targetNode=(ReferenceTarget*)rtarg;
			ICustButton *iBut = GetICustButton(GetDlgItem(hwnd2,IDC_BUTTON_TARGET));
			ICustEdit *iEdit = GetICustEdit(GetDlgItem(hwnd1,IDC_EDIT_IKFKSWIVEL));
			ISpinnerControl *iSpinner=GetISpinner(GetDlgItem(hwnd1,IDC_SPIN_IKFKSWIVEL));
			
			if (iBut && iEdit && iSpinner)
			{
				INode *targetNode=GetTargetNode();
				if (targetNode) 
					iBut->SetText(targetNode->GetName());
				else
					iBut->SetText(_T("None"));
				iEdit->Enable(!targetNode);
				iSpinner->Enable(!targetNode);
				ReleaseISpinner(iSpinner);
				ReleaseICustEdit(iEdit);
				ReleaseICustButton(iBut);
			}
		}
		break;
	case MAIN_REF:
		mainNode=(ReferenceTarget*)rtarg;
		break;
	case SEC_REF:
		secNode=(ReferenceTarget*)rtarg;
		break;
	case LAST_REF:
		lastNode=(ReferenceTarget*)rtarg;
		break;
	case NONUSED_REF:
		return;
		break;
	default:
		if (pblock && i>0 && i<(NSOLVER_REFS+GoalParentCount()))
			goalParentNode[i-NSOLVER_REFS]=(ReferenceTarget*)rtarg;
		break;
	}
}
int IKFKSolver::NumSubs()
{
	return SOLVER_NSUBS;
}
Animatable* IKFKSolver::SubAnim(int i)
{
	return pblock;
}
TSTR IKFKSolver::SubAnimName(int i)
{
	return _T("Parameters");
}


//Set and Get Nodes functions
//these functions sets or gets the nodes of the system having in mind that these are weak references
//that means that the controllers of the nodes can have a reference of this solver without making 
//dependency loops, when a node is added to the solver a fk or ik controler has to be assigned to it

INode *IKFKSolver::GetGoalNode()
{
	if (goalNode)
		return ((INodeTransformMonitor*)goalNode->GetInterface(IID_NODETRANSFORMMONITOR))->GetNode();
	return NULL;
}
INode *IKFKSolver::GetTargetNode()
{
	if (targetNode)
		return ((INodeTransformMonitor*)targetNode->GetInterface(IID_NODETRANSFORMMONITOR))->GetNode();
	return NULL;
}
INode *IKFKSolver::GetMainNode()
{
	if (mainNode)
		return ((INodeTransformMonitor*)mainNode->GetInterface(IID_NODETRANSFORMMONITOR))->GetNode();
	return NULL;
}
INode *IKFKSolver::GetSecNode()
{
	if (secNode)
		return ((INodeTransformMonitor*)secNode->GetInterface(IID_NODETRANSFORMMONITOR))->GetNode();
	return NULL;
}
INode *IKFKSolver::GetLastNode()
{
	if (lastNode)
		return ((INodeTransformMonitor*)lastNode->GetInterface(IID_NODETRANSFORMMONITOR))->GetNode();
	return NULL;
}

INode *IKFKSolver::GetGoalParentNode(int index)
{
	if (index<goalParentNode.Count() && goalParentNode[index])
		return ((INodeTransformMonitor*)goalParentNode[index]->GetInterface(IID_NODETRANSFORMMONITOR))->GetNode();
	return NULL;
}
void IKFKSolver::SetGoalNode(INode *theNode)
{
	
	//remove previous node ikcontrol
	INode *prevNode=GetGoalNode();
	
	//set reference, if theNode is null it means that the solver is being deleted, in that case the whole 
	//node monitor reference must be deleted to avoid crash
	if (theNode)
	{
		if (goalNode)
			((INodeTransformMonitor*)goalNode->GetInterface(IID_NODETRANSFORMMONITOR))->SetNode(theNode);
	}
	else
		ReplaceReference(GOAL_REF,NULL);
	if (prevNode)
	{
		Control* theCtrl=prevNode->GetTMController();
		if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
		{
			BOOL update=CanUpdate();
			if (update) 
				SetCanUpdate(FALSE);
			Control *theIK=((IKCtrl*)theCtrl)->IK;
			if (theIK)
				SetTMController(prevNode,theIK);
			else
				SetTMController(prevNode,NewDefaultMatrix3Controller());
			if (update)
				SetCanUpdate(TRUE);
		}
	}

	if (theNode)	//assign the ikcontrol
	{
		IKCtrl *newIKControl=new IKCtrl(this);
		newIKControl->Copy(theNode->GetTMController());
		SetTMController(theNode,newIKControl);
	}
}
void IKFKSolver::SetTargetNode(INode *theNode)
{
	BOOL isNotHolding=!theHold.Holding();
	if (isNotHolding)
		theHold.Begin();
	INode *prevNode=GetTargetNode();
	TimeValue t=GetCOREInterface()->GetTime();
	SetCanUpdate(FALSE);
	if (!theNode)
	{
		Solve(t,FORCE_SOLVE);
		SwivelMatch(t);
	}
	ReplaceReference(TARGET_REF,NULL);

	if (theNode)	//assign the ikcontrol
	{
		Match(t);
		ReplaceReference(TARGET_REF,(ReferenceTarget*)CreateInstance(REF_TARGET_CLASS_ID, NODETRANSFORMMONITOR_CLASS_ID));
		IKCtrl *newIKControl=new IKCtrl(this);
		newIKControl->Copy(theNode->GetTMController());
		SetTMController(theNode,newIKControl);

		//set reference
		if (targetNode)
			((INodeTransformMonitor*)targetNode->GetInterface(IID_NODETRANSFORMMONITOR))->SetNode(theNode);
		Match(t);
	}
	
	//remove previous node ikcontrol
	if (prevNode)
	{
		Control* theCtrl=prevNode->GetTMController();
		if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
		{
			Control *theIK=((IKCtrl*)theCtrl)->IK;
			if (theIK)
				SetTMController(prevNode,theIK);
			else
				SetTMController(prevNode,NewDefaultMatrix3Controller());
		}
	}
	SetCanUpdate(TRUE);
	NotifyChange();
	if (isNotHolding)
		theHold.Accept(_T("Change IKFK target"));
}
void IKFKSolver::SetMainNode(INode *theNode)
{
	//remove previous node fkcontrol
	
	INode *prevNode=GetMainNode();
	//set reference, if theNode is null it means that the solver is being deleted, in that case the whole 
	//node monitor reference must be deleted to avoid crash
	if (theNode)
	{
		if (mainNode)
			((INodeTransformMonitor*)mainNode->GetInterface(IID_NODETRANSFORMMONITOR))->SetNode(theNode);
	}
	else
		ReplaceReference(MAIN_REF,NULL);
	if (prevNode)
	{
		Control *theCtrl=prevNode->GetTMController();
		if (theCtrl->ClassID()==FKCtrl_CLASS_ID)
		{
			BOOL update=CanUpdate();
			if (update) 
				SetCanUpdate(FALSE);
			Control *theFK=((FKCtrl*)theCtrl)->FK;
			if (theFK)
				SetTMController(prevNode,theFK);
			else
				SetTMController(prevNode,NewDefaultMatrix3Controller());
			if (update)
				SetCanUpdate(TRUE);
		}
	}
	if (theNode)	//assign the ikcontrol
	{
		FKCtrl *newFKControl=new FKCtrl(this);
		newFKControl->Copy(theNode->GetTMController());
		SetTMController(theNode,newFKControl);
	}	
}
void IKFKSolver::SetSecNode(INode *theNode)
{
	
	//remove previous node fkcontrol
	INode *prevNode=GetSecNode();
	
	//set reference, if theNode is null it means that the solver is being deleted, in that case the whole 
	//node monitor reference must be deleted to avoid crash
	if (theNode)
	{
		if (secNode)
			((INodeTransformMonitor*)secNode->GetInterface(IID_NODETRANSFORMMONITOR))->SetNode(theNode);
	}
	else
		ReplaceReference(SEC_REF,NULL);
	if (prevNode)
	{
		Control *theCtrl=prevNode->GetTMController();
		if (theCtrl->ClassID()==FKCtrl_CLASS_ID)
		{
			BOOL update=CanUpdate();
			if (update) 
				SetCanUpdate(FALSE);
			Control *theFK=((FKCtrl*)theCtrl)->FK;
			if (theFK)
				SetTMController(prevNode,theFK);
			else
				SetTMController(prevNode,NewDefaultMatrix3Controller());
			if (update)
				SetCanUpdate(TRUE);
		}
	}

	if (theNode)	//assign the ikcontrol
	{
		FKCtrl *newFKControl=new FKCtrl(this);
		newFKControl->Copy(theNode->GetTMController());
		SetTMController(theNode,newFKControl);
	}
}
void IKFKSolver::SetLastNode(INode *theNode)
{
	
	//remove previous node fkcontrol
	INode *prevNode=GetLastNode();
	
	//set reference, if theNode is null it means that the solver is being deleted, in that case the whole 
	//node monitor reference must be deleted to avoid crash
	if (theNode)
	{
		if (lastNode)
			((INodeTransformMonitor*)lastNode->GetInterface(IID_NODETRANSFORMMONITOR))->SetNode(theNode);
	}
	else
		ReplaceReference(LAST_REF,NULL);
	if (prevNode)
	{
		Control *theCtrl=prevNode->GetTMController();
		if (theCtrl->ClassID()==FKCtrl_CLASS_ID)
		{
			BOOL update=CanUpdate();
			if (update) 
				SetCanUpdate(FALSE);
			Control *theFK=((FKCtrl*)theCtrl)->FK;
			if (theFK)
				SetTMController(prevNode,theFK);
			else
				SetTMController(prevNode,NewDefaultMatrix3Controller());
			if (update)
				SetCanUpdate(TRUE);
		}
	}
	if (theNode)	//assign the ikcontrol
	{
		FKCtrl *newFKControl=new FKCtrl(this);
		newFKControl->Copy(theNode->GetTMController());
		SetTMController(theNode,newFKControl);
	}
}

void IKFKSolver::SetGoalParentNode(INode *theNode)
{
	BOOL isNotHolding=!theHold.Holding();
	if (isNotHolding)
		theHold.Begin();
	SetCanUpdate(FALSE);
	int i,j,n,prevn=0;
	
	prevn=GoalParentCount();
	//remove previous parent nodes and ik controls
	//SetGoalParentCount(0);
	for (i=0; i<prevn; i++)
	{
		INode *parentNode=GetGoalParentNode(i);
		if (parentNode)
		{	
			Control* theCtrl=parentNode->GetTMController();
			if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
			{
				Control *theIK=((IKCtrl*)theCtrl)->IK;
				if (theIK)
					SetTMController(parentNode,theIK);
				else
					SetTMController(parentNode,NewDefaultMatrix3Controller());
			}
		}
	}
	if (theNode)
	{
		Tab<INode*>theNodes;
		INode *aux,*goalNode;
		//get the goal parents node list
		goalNode=GetGoalNode();
		aux=goalNode->GetParentNode();
		BOOL notFound=TRUE;
		while(aux && notFound)
		{
			if (aux==theNode) notFound=FALSE;
			theNodes.Append(1,&aux);
			aux=aux->GetParentNode();
		}
		if (!notFound)
		{
			n=theNodes.Count();
			//add new node monitors, new nodes and controls
			SetGoalParentCount(n);
			for (i=0; i<n; i++)
			{
				j=n-i-1;
				SetGoalParentNode(theNodes[j],i);
			}
		}
		
	}
	else
	{
		SetGoalParentCount(0);
	}
	SetCanUpdate(TRUE);	
	if (isNotHolding)
		theHold.Accept(_T("Change IKFK goal Parent"));
}
void IKFKSolver::SetGoalParentNode(INode *theNode, int index)
{
	if (index<GoalParentCount())
	{
		if (theNode)	//assign the ikcontrol
		{
			Control *theCtrl=theNode->GetTMController();
			IKCtrl *newIKControl;
			newIKControl=new IKCtrl(this);
			newIKControl->Copy(theCtrl);
			SetTMController(theNode,newIKControl);
			//set reference
			if (goalParentNode[index])
				((INodeTransformMonitor*)goalParentNode[index]->GetInterface(IID_NODETRANSFORMMONITOR))->SetNode(theNode);
		}
	}
}
void IKFKSolver::SetGoalParentCount(int n)
{
	if (pblock)
	{
		int prevn=goalParentNode.Count();		
		int i;
		pblock->SetValue(pb_nGoalParent,0,n);
		if (n>prevn)
			goalParentNode.SetCount(n);
		for (i=prevn; i<n; i++)
		{
			goalParentNode[i]=NULL;	
			ReplaceReference(NSOLVER_REFS+i,(ReferenceTarget*)CreateInstance(REF_TARGET_CLASS_ID, NODETRANSFORMMONITOR_CLASS_ID));
		}
	}
}


/*
FUNCTION IKFKSolver::CalculateIK
Job:
	Calculates the rotation of the fk objects of the ikfk system, 
Parameters:
	TimeValue t:
		the calculation time
	Interval &valid:
		the time Interval
	unsigned char solveFlags:
		flags for specific solving,
		FORCE_SOLVE (1<<0) - The solve will be calculated even if the change flag is false
		FULL_IKGOAL (1<<1) - calculated in full ik
		FULL_IKTARGET (1<<2) - calculated in ik but the goal position will be the current one
		FULL_FKMAIN (1<<3) - calculated with the main node in fk
		FULL_FKSEC (1<<4) - calculated with the sec node in fk
*/
void IKFKSolver::CalculateIK(TimeValue t, Interval &valid,  unsigned char solveFlags)
{
	INode *goalNode,*mainNode,*secNode,*lastNode;
	goalNode=GetGoalNode();
	mainNode=GetMainNode();
	secNode=GetSecNode();
	lastNode=GetLastNode();
	if (goalNode && mainNode && secNode && lastNode && pblock)
	{
		int parentSpace, nGoalParent;
		INode *targetNode,*goalParentNode;
		float IKFKBlend,trueIKFKBlend,scaleMult,IKStretch;
		Quat mainIKRot,secIKRot;
		IParamBlock2 *thePblock;
		Control *theCtrl;

		//values used to store position, rotation and scale of FK nodes (local space)
		Matrix3 FKParentTM,scaleTM,goalParentTM,solverParentTM;
		Point3 mainPos,secPos,lastPos;
		Quat lastRot; //mainRot and secRot are now IKFKSolver local variables
		Point3 mainScale,secScale,lastScale;

		
		targetNode=GetTargetNode();
		goalParentNode=GetGoalParentNode(0);
		nGoalParent=GoalParentCount();	

		pblock->GetValue(pb_IKFKBlend,t,IKFKBlend,valid);	
		pblock->GetValue(pb_parentSpace,t,parentSpace,FOREVER);
		pblock->GetValue(pb_scaleMult,t,scaleMult,valid);
		pblock->GetValue(pb_IKStretch,t,IKStretch,valid);

		trueIKFKBlend=IKFKBlend;
		if ((solveFlags & FULL_IKGOAL) || (solveFlags & FULL_IKTARGET))//if the chain is being manipulated in IK
			IKFKBlend=1.0;
		else if ((solveFlags & FULL_FKMAIN) && (solveFlags & FULL_FKSEC)) //if both full FK in main and Sec then no IK at all
			IKFKBlend=0.0;

		
		//get the transforms of the nodes, ik nodes and the first fk node in world space, the rest in parent

		if (!(solveFlags & FULL_IKTARGET)) //if the solve is in full target the goal tms will be taken from the previous calculation
		{
			if (nGoalParent && goalParentNode)
			{
				solvedNumGoalParent=GoalParentCount();
				gParentTM= new Matrix3[solvedNumGoalParent+1];
				Matrix3 prevTM=GetParentTM(t,goalParentNode,valid);
				//the rest of parent nodes tms will calculated without the first parent tm
				for (int i=0;i<nGoalParent;i++)
				{
					goalParentNode=GetGoalParentNode(i);
					if (goalParentNode)
					{
						theCtrl=goalParentNode->GetTMController();
						//it is suposed that this node will have always an IK/FK control, but this check is added for stability reasons.
						if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
							if (((IKCtrl*)theCtrl)->IK)
								((IKCtrl*)theCtrl)->IK->GetValue(t,&prevTM,valid,CTRL_RELATIVE);
						else
							theCtrl->GetValue(t,&prevTM,valid,CTRL_RELATIVE);
					}
					gParentTM[i]=prevTM;
				}

				goalParentTM=prevTM;
				theCtrl=goalNode->GetTMController();
				//it is suposed that this node will have always an IK/FK control, but this check is added for stability reasons.
				if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
					if (((IKCtrl*)theCtrl)->IK)
						((IKCtrl*)theCtrl)->IK->GetValue(t,&prevTM,valid,CTRL_RELATIVE);
				else
					theCtrl->GetValue(t,&prevTM,valid,CTRL_RELATIVE);
				goalTM=gParentTM[nGoalParent]=prevTM;			
			}
			else
			{
				goalTM=goalParentTM=GetParentTM(t,goalNode,valid);
				theCtrl=goalNode->GetTMController();
				//it is suposed that this node will have always an IK/FK control, but this check is added for stability reasons.
				if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
					if (((IKCtrl*)theCtrl)->IK)
						((IKCtrl*)theCtrl)->IK->GetValue(t,&goalTM,valid,CTRL_RELATIVE);
				else
					theCtrl->GetValue(t,&goalTM,valid,CTRL_RELATIVE);
			}
		}
		
		if (parentSpace)
		{
			solverParentTM=GetParentTM(t,mainNode,valid);
		}
		else
			solverParentTM=goalParentTM;
		if (targetNode)
		{
			targetTM=GetTargetParentTM(t,targetNode,valid);
			theCtrl=targetNode->GetTMController();
			//it is suposed that this node will have always an IK/FK control, but this check is added for stability reasons.
			if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
				if (((IKCtrl*)theCtrl)->IK)
					((IKCtrl*)theCtrl)->IK->GetValue(t,&targetTM,valid,CTRL_RELATIVE);
			else
				theCtrl->GetValue(t,&targetTM,valid,CTRL_RELATIVE);
		}

		//get the pos,rot and scale values of the main, sec and last nodes
		FKParentTM=GetParentTM(t,mainNode,valid);
		theCtrl=mainNode->GetTMController();
		//it is suposed that these nodes will have always an IK/FK control, but these checks are added for stability reasons.
		if (theCtrl->ClassID()==FKCtrl_CLASS_ID)
			theCtrl=((FKCtrl*)theCtrl)->FK;
		GetCtrlPRS(t,theCtrl,mainPos,mainRot,mainScale,valid);
		theCtrl=secNode->GetTMController();
		if (theCtrl->ClassID()==FKCtrl_CLASS_ID)
			theCtrl=((FKCtrl*)theCtrl)->FK;
		GetCtrlPRS(t,theCtrl,secPos,secRot,secScale,valid);
		theCtrl=lastNode->GetTMController();
		if (theCtrl->ClassID()==FKCtrl_CLASS_ID)
			theCtrl=((FKCtrl*)theCtrl)->FK;
		GetCtrlPRS(t,theCtrl,lastPos,lastRot,lastScale,valid);

		//now find total scale per node by multiplyng scale vectors
		mainScale*=GetTMScale(FKParentTM);
		secScale*=mainScale;
		lastScale*=secScale;
		//multiply positions by parent scales
		mainPos=mainPos*FKParentTM; //main position in world space
		secPos*=mainScale; 
		lastPos*=secScale;
		//multiply sec and last position by scalemult
		secPos*=scaleMult;
		lastPos*=scaleMult;
		//the main rot is in world space
		mainRot*=MatrixToQuat(FKParentTM);
		secRot*=mainRot;
	
		if (IKFKBlend) 
		{	
			Point3 goalPos,secIKPos;
			float GMdist,secdist,lastdist;
			Quat lookat;
			Matrix3 lookattm;

			//some vectors
			goalPos=goalTM.GetTrans();
			
			//some distances
			secdist=secPos.FLength();
			lastdist=lastPos.FLength();
			GMdist=(goalPos-mainPos).FLength();
			
			//the secIKPos is calculated
			if (GMdist<(secdist+lastdist))
			{
				if (GMdist<(lastdist-secdist))
					secIKPos=Point3(-secdist,0.0f,0.0f);
				else if (GMdist<(secdist-lastdist))
					secIKPos=Point3(secdist,0.0f,0.0f);
				else
				{	
					secIKPos.x=(secdist*secdist-lastdist*lastdist+GMdist*GMdist)/(2.0*GMdist);
					secIKPos.y=-Sqrt(secdist*secdist-secIKPos.x*secIKPos.x);
					secIKPos.z=0.0;
				}
			}
			else
			{
				secIKPos=Point3(secdist,0.0f,0.0f);
				//stretching
				float stretchMult;
				stretchMult=GMdist/(secdist+lastdist);
				stretchMult=1.0f+(stretchMult-1.0f)*IKStretch*trueIKFKBlend;
				secPos*=stretchMult;
				lastPos*=stretchMult;
			}
			//lookat orientation
			if (targetNode)//lookat with target as up vector
				lookattm=MakeLookAtMatrix(mainPos,goalPos,targetTM.GetTrans());
			else//use swivel and  lookat with parenttm (solverParentTM) as up vector
			{
				float mainPrefRot[3];
				Matrix3 mainPrefTM;
				//use the main preferred rotation as parent as well
				thePblock=mainNode->GetTMController()->GetParamBlock(LOCALPBLOCK_REF);
				mainPrefRot[0]=0.0f;
				mainPrefRot[1]=0.0f;
				mainPrefRot[2]=0.0f;
				if (thePblock)
				{
					thePblock->GetValue(0,t,mainPrefRot[0],valid);
					thePblock->GetValue(1,t,mainPrefRot[1],valid);
					thePblock->GetValue(2,t,mainPrefRot[2],valid);
				}
				EulerToMatrix(mainPrefRot,mainPrefTM,EULERTYPE_XYZ);
				float swivelAngle;
				pblock->GetValue(pb_IKFKSwivelAngle,t,swivelAngle,valid);
				lookat=Quat(AngAxis(Point3(1.0f,0.0f,0.0f),swivelAngle))*MakeLookAtQuat(mainPos,goalPos,mainPrefTM*solverParentTM);
				lookat.MakeMatrix(lookattm);
				lookattm.SetTrans(mainPos);
			}	

			//secPrefRots
			float secPrefRot[3];
			thePblock=secNode->GetTMController()->GetParamBlock(LOCALPBLOCK_REF);
			secPrefRot[0]=0.0f;
			secPrefRot[1]=0.0f;
			secPrefRot[2]=0.0f;
			if (thePblock)
			{
				thePblock->GetValue(0,t,secPrefRot[0],valid);
				thePblock->GetValue(1,t,secPrefRot[1],valid);
				thePblock->GetValue(2,t,secPrefRot[2],valid);
			}
			Matrix3 secPreftm;
			EulerToMatrix(secPrefRot,secPreftm,EULERTYPE_XYZ);
			

			//the lookat tms for the main and sec node are calculated here, they match the position of the main, goal 
			//and the sec IK pos calcualted earlier
			Point3 secIKPos2=secIKPos*lookattm;
			Matrix3 mainIKtm;		
			Point3 mainUpVector=lastPos*secPreftm;
			mainUpVector.y*=-1;
			Matrix3 prefLooktm=MakeLookAtMatrix(Point3(0.0f,0.0f,0.0f),secPos,mainUpVector)*lookattm;			
			mainIKRot=RotVector(secPos,secIKPos2*Inverse(prefLooktm))*Quat(prefLooktm);
			mainIKRot.MakeMatrix(mainIKtm);
			mainIKtm.SetTrans(secIKPos2);
			secIKRot=RotVector(lastPos,goalPos*Inverse(secPreftm*mainIKtm))*Quat(secPreftm)*mainIKRot;
			
			//the closest rotation to fk is calculated
			mainIKRot.MakeClosest(mainRot);
			secIKRot.MakeClosest(secRot);

			// if full_fkmain or full_fksec is on the second rotation has to be calculated in local space
			if (solveFlags & (FULL_FKMAIN | FULL_FKSEC))
			{
				secRot=secRot*Inverse(mainRot);
				secIKRot=secIKRot*Inverse(mainIKRot);
			}

			//IKFK blending
			if (IKFKBlend<1) //avoid using slerp when not necesary
			{
				if (!(solveFlags & FULL_FKMAIN)) 
					mainRot=Slerp(mainRot,mainIKRot,IKFKBlend);
				if (!(solveFlags & FULL_FKSEC))
					secRot=Slerp(secRot,secIKRot,IKFKBlend);
			}
			else
			{
				if (!(solveFlags & FULL_FKMAIN))
					mainRot=mainIKRot;
				if (!(solveFlags & FULL_FKSEC))
					secRot=secIKRot;
			}

			if (solveFlags & (FULL_FKMAIN | FULL_FKSEC))
				secRot=secRot*mainRot;	
			
		}
		//Sets World FK Nodes TMs
		//first set rot and pos		
		mainRot.MakeMatrix(mainTM);
		mainTM.SetTrans(mainPos);
		secRot.MakeMatrix(secTM);
		secTM.SetTrans(secPos*mainTM);
		lastRot*=secRot;//lastrot was in local space
		lastRot.MakeMatrix(lastTM);
		lastTM.SetTrans(lastPos*secTM);
		//then set scales (the scales are already in world space)
		
		scaleTM.SetScale(mainScale);
		mainTM=scaleTM*mainTM;
		scaleTM.SetScale(secScale);
		secTM=scaleTM*secTM;
		scaleTM.SetScale(lastScale);
		lastTM=scaleTM*lastTM;


		goalTM.SetTrans(lastTM.GetTrans()); //setting Goal TM, it's position is the same as the last node's one
		
		//-----------------------------------------------------------------------
		//Calculate goal parent position		
		//(Inverse ((Inverse (CSR*BSR*ASR*TP))*CSR*CP*BSR*BP*ASR)).position;
		if (nGoalParent)
		{
			Matrix3 goalTM=gParentTM[nGoalParent];
			Matrix3 goalTMnoParentTrans=gParentTM[0];
			goalTM.SetTrans(lastTM.GetTrans());
			goalTMnoParentTrans.NoTrans();
			goalTMnoParentTrans=gParentTM[nGoalParent]*Inverse(gParentTM[0])*goalTMnoParentTrans;
			Point3 goalParentTrans=(Inverse(Inverse(goalTM)*goalTMnoParentTrans)).GetTrans()-gParentTM[0].GetTrans();
			for (int i=0;i<nGoalParent;i++)
				gParentTM[i].Translate(goalParentTrans);
		}
	}
}
/*
FUNCTION IKFKSolver::Solve
Job:
	Calculates the transform of the objects of the ikfk system, the values are stored in the "value" variable
	of the fkcntrl and ikcntrl
Parameters:
	TimeValue t:
		time of calculation
	unsigned char solveFlags:
		flags for specific solving,
		FORCE_SOLVE (1<<0) - The solve will be calculated even if the change flag is false
		FULL_IKGOAL (1<<1) - calculated in full ik
		FULL_IKTARGET (1<<2) - calculated in ik but the goal position will be the current one
		FULL_FKMAIN (1<<3) - calculated with the main node in fk
		FULL_FKSEC (1<<4) - calculated with the sec node in fk
Returns:
	returns false if the Solver is already processing, true otherwise
*/
BOOL IKFKSolver::Solve(TimeValue t, unsigned char solveFlags)
{
	//the bool notProcessing is used to avoid crashes when there is a dependency loop
	if (t!=prevt || hasChanged || (solveFlags & FORCE_SOLVE))
	{
		if (notProcessing)
		{
			notProcessing=FALSE;
			INode *theNode;
			Interval valid=FOREVER;
			Control *theCtrl;
			//if solve is in full target another solve has to be made to get the current goal and goalparent 
			//tms and use it for the next solve
			if (solveFlags & FULL_IKTARGET)
				CalculateIK(t,valid,0);

			//solve the ik value
			CalculateIK(t,valid,solveFlags);

			//set interval values
			theNode=GetTargetNode();
			if (theNode)
			{
				theCtrl=theNode->GetTMController();
				if (theCtrl->ClassID()==IKCtrl_CLASS_ID)//check for stability reasons
				{
					((IKCtrl*)theCtrl)->value=targetTM;
					((IKCtrl*)theCtrl)->ivalid=valid;
				}
			}
			//set goal parent nodes tm
			for (int i=0;((i<GoalParentCount()) && (i<solvedNumGoalParent)) ;i++)
			{
				theNode=GetGoalParentNode(i);
				if (theNode)
				{
					theCtrl=theNode->GetTMController();
					if (theCtrl->ClassID()==IKCtrl_CLASS_ID)//check for stability reasons
					{
						((IKCtrl*)theCtrl)->value=gParentTM[i];
						((IKCtrl*)theCtrl)->ivalid=valid;
					}
				}
			}
			theNode=GetGoalNode();
			if (theNode)
			{
				theCtrl=theNode->GetTMController();
				if (theCtrl->ClassID()==IKCtrl_CLASS_ID)//check for stability reasons
				{
					((IKCtrl*)theCtrl)->value=goalTM;
					((IKCtrl*)theCtrl)->ivalid=valid;
				}
			}

			theNode=GetMainNode();
			if (theNode)
			{
				theCtrl=theNode->GetTMController();
				if (theCtrl->ClassID()==FKCtrl_CLASS_ID)//check for stability reasons
				{
					((FKCtrl*)theCtrl)->value=mainTM;
					((FKCtrl*)theCtrl)->ivalid=valid;
				}
			}

			theNode=GetSecNode();
			if (theNode)
			{
				theCtrl=theNode->GetTMController();
				if (theCtrl->ClassID()==FKCtrl_CLASS_ID)//check for stability reasons
				{
					((FKCtrl*)theCtrl)->value=secTM;
					((FKCtrl*)theCtrl)->ivalid=valid;
				}
			}

			theNode=GetLastNode();
			if (theNode)
			{
				theCtrl=theNode->GetTMController();
				if (theCtrl->ClassID()==FKCtrl_CLASS_ID)//check for stability reasons
				{
					((FKCtrl*)theCtrl)->value=lastTM;
					((FKCtrl*)theCtrl)->ivalid=valid;
				}
			}
			prevt=t;
			hasChanged=FALSE;
			notProcessing=TRUE;
			return TRUE;
		}
		else
			return FALSE;
	}
	return TRUE;
}

//Match
//this matches both ik and fk to the current values (that depends on the current ikfk blending 
//and ik and fk tms), the target, swivel, scale mult, mainprefrot and secprefrot are matched
//as well (if their respective flags are on).
void IKFKSolver::Match(TimeValue t)
{
	SetCanUpdate(FALSE);
	BOOL matchScaleMult=FALSE,matchMainPref=FALSE,matchSecPref=FALSE;
	INode *targetNode=GetTargetNode();
	if (pblock)
	{
		pblock->GetValue(pb_matchScaleMult,0,matchScaleMult,FOREVER);
		pblock->GetValue(pb_matchMainPreferred,0,matchMainPref,FOREVER);
		pblock->GetValue(pb_matchSecPreferred,0,matchSecPref,FOREVER);
	}
	//the order of matching here is important!!!!, the target and swivel matches depends on wether or not the
	//sec pref rot has changed
	Solve(t,FORCE_SOLVE);
	if (matchSecPref)
		SecPrefMatch(t);
	if (matchScaleMult)
		ScaleMultMatch(t);
	FKMatch(t);
	IKMatch(t);
	if (targetNode)
		TargetMatch(t);
	else
		SwivelMatch(t);
	if (matchMainPref)
		MainPrefMatch(t);
	SetCanUpdate(TRUE);
	NotifyChange();
}

//FKMatch
//the FKmatch sets the local fk goal rotation value to the one that is currently representing
//with its "FK->value" variable, it has as parameter the controller that has been transformed
//to make the match only in that node (we dont want rotations keys created in a fk node that we
//have not rotated)
// It uses the mainRot and secRot values instead of using directly the tm value, so it can avoid 
// odd scale transformation problems
void IKFKSolver::FKMatch(TimeValue t,Control *theCtrl)
{
	INode *mainNode,*secNode;

	mainNode=GetMainNode();
	secNode=GetSecNode();
	if (mainNode && secNode)
	{
		Control *mainCtrl,*secCtrl;
		mainCtrl=mainNode->GetTMController();
		secCtrl=secNode->GetTMController();
		if(mainCtrl->ClassID()==FKCtrl_CLASS_ID && secCtrl->ClassID()==FKCtrl_CLASS_ID)//check for stability reasons
		{
			Matrix3 parenttm;
			Quat mainRot,secRot;
			Interval valid=FOREVER;

			//match main node
			if (!theCtrl || theCtrl==mainCtrl)
			{
				Control *theTMCtrl=((FKCtrl*)mainCtrl)->FK;
				if (theTMCtrl)
				{
					parenttm=GetParentTM(t,mainNode,FOREVER);
					//if the TM control has a rotation control, the match has to be done there directly
					Control *theRotCtrl=theTMCtrl->GetRotationController();
					if (theRotCtrl)
					{
						//mainRot was in global space, has to be converted to local space
						mainRot=this->mainRot;
						mainRot*=Inverse(MatrixToQuat(parenttm));
						theRotCtrl->SetValue(t,&mainRot,1,CTRL_ABSOLUTE);
					}
					else
						theTMCtrl->SetValue(t,&SetXFormPacket(mainTM,parenttm),1,CTRL_ABSOLUTE);
				}
			}
			//match sec node
			if (!theCtrl || theCtrl==secCtrl)
			{
				Control *theTMCtrl=((FKCtrl*)secCtrl)->FK;
				if (theTMCtrl)
				{
					//if the TM control has a rotation control, the match has to be done there directly
					Control *theRotCtrl=theTMCtrl->GetRotationController();
					if (theRotCtrl)
					{
						//secRot was in global space, has to be converted to local space
						secRot=this->secRot;
						secRot*=Inverse(this->mainRot);
						theRotCtrl->SetValue(t,&secRot,1,CTRL_ABSOLUTE);
					}
					else
						theTMCtrl->SetValue(t,&SetXFormPacket(secTM,mainTM),1,CTRL_ABSOLUTE);
				}
			}
		}
	}
}

//IKMatch
//the IKmatch sets the local ik goal position value to the one that is currently representing
//with its "IK->value" variable, if the solver has a goal parent the goal parent has to be used
//in this operation instead
void IKFKSolver::IKMatch(TimeValue t)
{
	Interval valid=FOREVER;
	INode *theIKNode;
	Matrix3 currenttm, parenttm, localtm;
	SetXFormPacket mat;
	Control *theCtrl;

	theIKNode=GetGoalParentNode(0);
	if (!theIKNode || !GoalParentCount())
	{
		theIKNode=GetGoalNode();
		currenttm=goalTM;
	}
	else if (solvedNumGoalParent)
		currenttm=gParentTM[0];
	else
		currenttm=Matrix3(1);

	//match IK Node
	if (theIKNode)
	{
		theCtrl=theIKNode->GetTMController();
		if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
		{
			parenttm=GetParentTM(t,theIKNode,FOREVER);
			localtm=parenttm;
			theCtrl=((IKCtrl*)theCtrl)->IK;

			if (theCtrl)
			{
				theCtrl->GetValue(t,&localtm,valid,(GetSetMethod)CTRL_RELATIVE);
				mat.command=XFORM_MOVE;
				mat.localOrigin=FALSE;
				mat.p=currenttm.GetTrans()-localtm.GetTrans();
				mat.tmAxis=Matrix3 (1);
				mat.tmParent=parenttm;
				if (theCtrl->ClassID()==AnimPivot_CLASS_ID || theCtrl->ClassID()==AnimPivotTarget_CLASS_ID)
					theCtrl=(Control*)(theCtrl->GetReference(ANIMPIVOT_TM_REF)->GetReference(1));
				if (theCtrl)
					theCtrl->SetValue(t,&mat,1,CTRL_RELATIVE);
			}
		}
	}
}


//Swivel Match
//in order to calculate the right swivel angle we have to keep in mind these facts:
//To find the nodes transformations the solver first find a lookat tm that is used to 
//define the orientation, it has its translation in the main node and is looking at
//the ik goal, the lookat is calculated building a quaternion that has the main pref rotation
//and the parent tm as lookup(defined in the parent space group, the parent of the IK Goal or
//the parent of the Start joint), both nodes, the main and the second node bend along the plane 
//defined by that lookat tm, and finally the swivel angle offsets the plane about the x axis of
//the lookat tm.
//The swivel angle has to be calculated using the position of the sec node with respect to the 
//lookat quat and the main node position
void IKFKSolver::SwivelMatch(TimeValue t)
{
	INode* mainNode=GetMainNode();
	INode* secNode=GetSecNode();
	INode* lastNode=GetLastNode();
	if (mainNode && secNode && lastNode && pblock)
	{
		Matrix3 parenttm, mainpreftm, looktm;
		float swivelAngle;
		float mainPrefRot[3];
		Point3 secpos2;
		Control *theCtrl;
		IParamBlock2 *thePblock;

		//get parenttm
		parenttm=GetParentTM(t,mainNode,FOREVER);

		//get mainpreftm
		thePblock=mainNode->GetTMController()->GetParamBlock(LOCALPBLOCK_REF);
		mainPrefRot[0]=0.0f;
		mainPrefRot[1]=0.0f;
		mainPrefRot[2]=0.0f;
		if (thePblock)
		{
			thePblock->GetValue(0,t,mainPrefRot[0],FOREVER);
			thePblock->GetValue(1,t,mainPrefRot[1],FOREVER);
			thePblock->GetValue(2,t,mainPrefRot[2],FOREVER);
		}
		EulerToMatrix(mainPrefRot,mainpreftm,EULERTYPE_XYZ);

		//find looktm
		(MakeLookAtQuat(mainTM.GetTrans(),lastTM.GetTrans(),mainpreftm*parenttm)).MakeMatrix(looktm);
		looktm.SetTrans(mainTM.GetTrans());

		float chainAngle=acos(DotProd(Normalize(lastTM.GetTrans()-mainTM.GetTrans()),Normalize(secTM.GetTrans()-mainTM.GetTrans())));

		// if the chain is extended a different calculation has to be made for the secpos2 using the sec 
		// pref rotation and the local last post.
		if (chainAngle<ANGLE_TRESHOLD)
		{
			//the secpreftm and lastPos are needed to find the upvector of the second look tm (useful when the chain is totally extended)
			float secPrefRot[3];
			Matrix3 secpreftm;
			thePblock=secNode->GetTMController()->GetParamBlock(LOCALPBLOCK_REF);
			secPrefRot[0]=0.0f;
			secPrefRot[1]=0.0f;
			secPrefRot[2]=0.0f;
			if (thePblock)
			{
				thePblock->GetValue(0,t,secPrefRot[0],FOREVER);
				thePblock->GetValue(1,t,secPrefRot[1],FOREVER);
				thePblock->GetValue(2,t,secPrefRot[2],FOREVER);
			}
			EulerToMatrix(secPrefRot,secpreftm,EULERTYPE_XYZ);
			Point3 lastPos;
			theCtrl=lastNode->GetTMController();
			Control *posCtrl=theCtrl->GetPositionController();
			if (posCtrl)
				posCtrl->GetValue(t,&lastPos,FOREVER,CTRL_ABSOLUTE);
			else
			{
				Matrix3 theTM=Matrix3(1);
				theCtrl->GetValue(t,&theTM,FOREVER,CTRL_RELATIVE);
				lastPos=theTM.GetTrans();
			}
			secpreftm*=mainTM;
			secpreftm.SetTrans(secTM.GetTrans());

			//find secpos2 (local secpos with respect to looktm and sec pref - lastpos)
			secpos2=Inverse(looktm).VectorTransform(secTM.GetTrans()-lastPos*secpreftm);
		}
		else
			secpos2=secTM.GetTrans()*Inverse(looktm);


		//find swivel angle
		pblock->GetValue(pb_IKFKSwivelAngle,t,swivelAngle,FOREVER); 
		//the previous swivel angle is used to find the closest value if it overpass 180 or -180 degrees
		swivelAngle=GetCloserAngle(swivelAngle,atan2(secpos2.z,-secpos2.y));

		//set new swivel angle	
		canUpdateSwivel=FALSE;
		pblock->SetValue(pb_IKFKSwivelAngle,t,swivelAngle);
		canUpdateSwivel=TRUE;
	}
}
//Target Match
// match tthe target using the TargetMatch1 and TargetMatch2 functions
void IKFKSolver::TargetMatch(TimeValue t)
{
	TargetMatch2(t,TargetMatch1(t,lastTM.GetTrans()));
}

//Target Match 1
// the target match is calculated using 2 lookat matrices, the first one is looking at the goal local position and uses the target's
// current position as lookup, the second one is looking at the last node and having the second node as lookup, both matrices have the main 
// node as origin. the target transform from one lookat matrix to the other.
// This is made in 2 functions, the first function calculates the new target position, the second function sets that new position in
// the target controllers.
// this is used this way to avoid double transformations when the target is linket to a goal parent that is part of the solver
Point3 IKFKSolver::TargetMatch1(TimeValue t, Point3 prevLastPos)
{
	INode *targetNode,*secNode,*lastNode,*goalNode;
	targetNode=GetTargetNode();
	secNode=GetSecNode();
	lastNode=GetSecNode();
	goalNode=GetGoalNode();

	if (secNode && lastNode && goalNode && targetNode)
	{
		Matrix3 goaltm,targettm,targetparenttm,looktm1,looktm2;
		Control *theCtrl;
		Control *theTargetCtrl=targetNode->GetTMController();

		if (theTargetCtrl->ClassID()==IKCtrl_CLASS_ID && ((IKCtrl*)theTargetCtrl)->IK) //check for stability reasons
			theCtrl=((IKCtrl*)theTargetCtrl)->IK;
		else
			theCtrl=theTargetCtrl;

		//get targettm
		targetparenttm=GetTargetParentTM(t,targetNode,FOREVER);
		targettm=targetparenttm;
		theCtrl->GetValue(t,&targettm,FOREVER,CTRL_RELATIVE);

		//get goaltm 
		theCtrl=goalNode->GetTMController();
		if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
		{
			theCtrl=((IKCtrl*)theCtrl)->IK;
			goaltm=GetGoalParentTM(t);
			theCtrl->GetValue(t,&goaltm,FOREVER,CTRL_RELATIVE);
		}
		else
			goaltm=goalTM;

		float chainAngle=acos(DotProd(Normalize(lastTM.GetTrans()-mainTM.GetTrans()),Normalize(secTM.GetTrans()-mainTM.GetTrans())));

		// if the chain is extended a different calculation has to be made for the lookat tms using the sec 
		// pref rotation and the local last post.
		if (chainAngle<ANGLE_TRESHOLD)
		{
			//the secpreftm and lastPos are needed to find the upvector of the second look tm (useful when the chain is totally extended)
			float secPrefRot[3];
			IParamBlock2 *thePblock;
			Matrix3 secpreftm;
			thePblock=secNode->GetTMController()->GetParamBlock(LOCALPBLOCK_REF);
			secPrefRot[0]=0.0f;
			secPrefRot[1]=0.0f;
			secPrefRot[2]=0.0f;
			if (thePblock)
			{
				thePblock->GetValue(0,t,secPrefRot[0],FOREVER);
				thePblock->GetValue(1,t,secPrefRot[1],FOREVER);
				thePblock->GetValue(2,t,secPrefRot[2],FOREVER);
			}
			EulerToMatrix(secPrefRot,secpreftm,EULERTYPE_XYZ);
			Point3 lastPos;
			theCtrl=lastNode->GetTMController();
			Control *posCtrl=theCtrl->GetPositionController();
			if (posCtrl)
				posCtrl->GetValue(t,&lastPos,FOREVER,CTRL_ABSOLUTE);
			else
			{
				Matrix3 theTM=Matrix3(1);
				theCtrl->GetValue(t,&theTM,FOREVER,CTRL_RELATIVE);
				lastPos=theTM.GetTrans();
			}
			secpreftm*=mainTM;
			secpreftm.SetTrans(secTM.GetTrans());

			//calculate lookat tms
			//the first tm is looking at the prevLastPos and has the target node as lookup vector
			looktm1=MakeLookAtMatrix(mainTM.GetTrans(),prevLastPos,targettm.GetTrans());
			//the second tm is looking at the last node and has the second node and the secpref-lastpos as lookup vector
			looktm2=MakeLookAtMatrix2((lastTM.GetTrans()-mainTM.GetTrans()),(lastPos*secpreftm-secTM.GetTrans()),mainTM.GetTrans());
		}
		else
		{
			//calculate lookat tms
			//the first tm is looking at the prevLastPos and has the target node as lookup vector
			looktm1=MakeLookAtMatrix(mainTM.GetTrans(),prevLastPos,targettm.GetTrans());
			//the second tm is looking at the last node and has the second node as lookup vector
			looktm2=MakeLookAtMatrix(mainTM.GetTrans(),lastTM.GetTrans(),secTM.GetTrans());
		}


		//set calculate and return the new target pos to the target local controller
		return ((targettm.GetTrans()*Inverse(looktm1))*looktm2);
	}
	return Point3(0.0f,0.0f,0.0f);
}
//Target Match
// this is the second function of the target match, where the previously calculated position is set in the target controller
void IKFKSolver::TargetMatch2(TimeValue t, Point3 newTargetPos)
{
	INode *targetNode;
	targetNode=GetTargetNode();

	if (targetNode)
	{
		Control *theTargetCtrl=targetNode->GetTMController();
		if (theTargetCtrl->ClassID()==IKCtrl_CLASS_ID && ((IKCtrl*)theTargetCtrl)->IK) //check for stability reasons
		{
			Matrix3 targettm,targetparenttm;
			Control *theCtrl;
			SetXFormPacket mat;

			//get targettm
			theCtrl=((IKCtrl*)theTargetCtrl)->IK;
			targetparenttm=GetTargetParentTM(t,targetNode,FOREVER);
			targettm=targetparenttm;
			theCtrl->GetValue(t,&targettm,FOREVER,CTRL_RELATIVE);


			//set calculate and set the new target pos to the target local controller
			mat.command=XFORM_MOVE;
			mat.p=newTargetPos-targettm.GetTrans();
			mat.tmAxis=Matrix3 (1);
			mat.tmParent=targetparenttm;
			mat.localOrigin=FALSE;
			((IKCtrl*)theTargetCtrl)->IK->SetValue(t,&mat,1,CTRL_RELATIVE);
		}
	}
}
//MainPrefMatch
//the main fk node preferred rotation is calculated by finding a lookat building a quaternion
//having the main node as origin and the last fk node as target
void IKFKSolver::MainPrefMatch(TimeValue t)
{
	INode *mainNode;

	mainNode=GetMainNode();
	
	if (mainNode)
	{
		IParamBlock2 *thePblock=mainNode->GetTMController()->GetParamBlock(LOCALPBLOCK_REF);	
		if (thePblock)
		{
			Interval valid=FOREVER;
			float mainPrefRot[3];
			int parentSpace=0;
			Matrix3 preftm,parenttm,lookattm;
			Quat prefquat,preftrans;
			Point3 vectorA,vectorB,mainpos,secpos,lastpos;

			if (pblock)
				pblock->GetValue(pb_parentSpace,t,parentSpace,valid);
			
			if (parentSpace)
				parenttm=GetParentTM(t,mainNode,valid);
			else
				parenttm=GetGoalParentTM(t);

			parenttm.Orthogonalize();
			parenttm.NoScale();


			mainpos=mainTM.GetTrans()*Inverse(parenttm);
			secpos=secTM.GetTrans()*Inverse(parenttm);
			lastpos=lastTM.GetTrans()*Inverse(parenttm);

			mainPrefRot[0]=0.0f;
			mainPrefRot[1]=0.0f;
			mainPrefRot[2]=0.0f;
			thePblock->GetValue(0,t,mainPrefRot[0],valid);
			thePblock->GetValue(1,t,mainPrefRot[1],valid);
			thePblock->GetValue(2,t,mainPrefRot[2],valid);
			EulerToMatrix(mainPrefRot,preftm,EULERTYPE_XYZ);

			//the new main pref rot is found by rotating its current value until it matches the 
			//orientation of the iklook quaternion
			prefquat=Quat(preftm);
			lookattm=MakeLookAtMatrix(mainpos,lastpos,secpos);
			vectorA=Normalize(lookattm.GetRow(0));
			vectorB=Normalize(preftm.GetRow(0));
			Quat newprefquat=prefquat*Quat(AngAxis(Normalize(CrossProd(vectorA,vectorB)),acos(DotProd(vectorA,vectorB))));
			newprefquat.MakeClosest(prefquat);

			QuatToEuler(newprefquat,mainPrefRot);
			thePblock->SetValue(0,t,mainPrefRot[0]);
			thePblock->SetValue(1,t,mainPrefRot[1]);
			thePblock->SetValue(2,t,mainPrefRot[2]);
		}
	}
}
//SecPrefMatch
//the sec fk node preferred rotation is matched to the current rotation of the second fk node
void IKFKSolver::SecPrefMatch(TimeValue t,BOOL forceMatch)
{
	INode *secNode=GetSecNode();
	if (secNode)
	{
		Quat secRot;
		IParamBlock2* thePblock;
		float secPrefRot[3];

		secRot=this->secRot*Inverse(this->mainRot);
		QuatToEuler(secRot,secPrefRot,EULERTYPE_XYZ);

		thePblock=secNode->GetTMController()->GetParamBlock(LOCALPBLOCK_REF);
		if (thePblock)
		{
			thePblock->SetValue(0,t,secPrefRot[0]);
			thePblock->SetValue(1,t,secPrefRot[1]);
			thePblock->SetValue(2,t,secPrefRot[2]);
		}
	}
}
//ScaleMultMatch
//the new scale value is found by taking the amount of current stretching and multiplying it
//by the current scale value
void IKFKSolver::ScaleMultMatch(TimeValue t,Control *theCtrl)
{
	if (pblock)
	{
		BOOL matchScaleMult;
		INode *goalNode,*mainNode,*secNode,*lastNode;
		pblock->GetValue(pb_matchScaleMult,t,matchScaleMult,FOREVER);
		goalNode=GetGoalNode();
		mainNode=GetMainNode();
		secNode=GetSecNode();
		lastNode=GetLastNode();

		if (matchScaleMult && goalNode && mainNode && secNode && lastNode && (!theCtrl || theCtrl==secNode->GetTMController()))
		{
			float scaleMult,IKStretch,GMdist,secdist,lastdist,stretchMult,IKFKBlend;
			Matrix3 goaltm,FKParentTM;
			Point3 mainPos,secPos,lastPos;
			Quat mainRot,secRot,lastRot;
			Point3 mainScale,secScale,lastScale;
			Control *theCtrl;

			pblock->GetValue(pb_scaleMult,t,scaleMult,FOREVER);
			pblock->GetValue(pb_IKStretch,t,IKStretch,FOREVER);
			pblock->GetValue(pb_IKFKBlend,t,IKFKBlend,FOREVER);

			//the transforms of the nodes are found by taking its local tms in the same way that
			//is done in the calculateik function

			//get the transforms of the nodes, ik nodes and the first fk node in world space, the rest in parent
			goaltm=GetGoalParentTM(t);
			theCtrl=goalNode->GetTMController();
			if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
				theCtrl=((IKCtrl*)theCtrl)->IK;
			if (theCtrl)
				theCtrl->GetValue(t,&goaltm,FOREVER,CTRL_RELATIVE);

			//get the pos,rot and scale values of the main, sec and last nodes
			FKParentTM=GetParentTM(t,mainNode,FOREVER);
			GetCtrlPRS(t,((Control*)mainNode->GetTMController()->GetReference(LOCALCTRL_REF)),mainPos,mainRot,mainScale,FOREVER);
			GetCtrlPRS(t,((Control*)secNode->GetTMController()->GetReference(LOCALCTRL_REF)),secPos,secRot,secScale,FOREVER);
			GetCtrlPRS(t,((Control*)lastNode->GetTMController()->GetReference(LOCALCTRL_REF)),lastPos,lastRot,lastScale,FOREVER);

			//now find total scale per node by multiplyng scale vectors
			mainScale*=GetTMScale(FKParentTM);
			secScale*=mainScale;
			//multiply positions by parent scales, sec and last position are limited to x axis
			mainPos=mainPos*FKParentTM; //main position in world space
			secPos*=mainScale; 
			lastPos*=secScale;
			//multiply sec and last position by scalemult
			secPos*=scaleMult;
			lastPos*=scaleMult;


			secdist=secPos.FLength();
			lastdist=lastPos.FLength();
			GMdist=(goaltm.GetTrans()-mainPos).FLength();

			if (GMdist>=(secdist+lastdist))
			{
				//stretching
				stretchMult=GMdist/(secdist+lastdist);
				stretchMult=1.0f+(stretchMult-1.0f)*IKStretch*IKFKBlend;
			}
			else
				stretchMult=1.0f;

			scaleMult=scaleMult*stretchMult;
			pblock->SetValue(pb_scaleMult,t,scaleMult);
		}
	}
}
//NumSelectedIK returns the number of selected IK nodes of the solver
int IKFKSolver::NumSelectedIKGoal()
{
	int numSelected=0;
	if (GetGoalNode()->Selected())
		numSelected++;
	for (int i=0;i<GoalParentCount();i++)
			if (GetGoalParentNode(i)->Selected())
				numSelected++;
	if (numSelected)
		numSelected=1;
	return numSelected;
}
//NumSelectedIK returns the number of selected IK nodes of the solver
int IKFKSolver::NumSelectedIKTarget()
{
	int numSelected=0;
	INode *target=GetTargetNode();
	if (target && target->Selected())
		numSelected=1;
	return numSelected;
}
//NumSelectedFK returns the number of selected FK nodes of the solver
int IKFKSolver::NumSelectedFK()
{
	int numSelected=0;
	if (GetMainNode()->Selected())
		numSelected++;
	if (GetSecNode()->Selected())
		numSelected++;
	if (GetLastNode()->Selected())
		numSelected++;
	if (numSelected)
		numSelected=1;
	return numSelected;
}

/*FUNCTION IKFKSolver::NotifyChange
Job:
	Calls the NotifyChange functions of every ikfk node controller,
	this is done to make sure that when one ikfk node notifies a change all the others have to notify the change as well
	the ikfk control that has been changed calls this function instead of notifying change.
Parameters:
	TimeValue t:
		time of calculation
*/
void IKFKSolver::NotifyChange()
{
	if (notProcessing)
	{
		notProcessing=FALSE;
		hasChanged=TRUE;
		INode *targetNode=GetTargetNode();
		INode *mainNode=GetMainNode();
		INode *secNode=GetSecNode();
		INode *lastNode=GetLastNode();
		INode *goalNode=GetGoalNode();
		INode *goalParentNode;
		int nGoalParent=GoalParentCount();
		if (targetNode)
			targetNode->GetTMController()->NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
		if (nGoalParent)
		{
			for (int i=0;i<nGoalParent;i++)
			{
				goalParentNode=GetGoalParentNode(i);
				if (goalParentNode)
					goalParentNode->GetTMController()->NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);	
			}
		}
		else if (goalNode)
			goalNode->GetTMController()->NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
		if (mainNode && secNode && lastNode)
		{
			mainNode->GetTMController()->NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
			secNode->GetTMController()->NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
			lastNode->GetTMController()->NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
		}
		notProcessing=TRUE;
	}
}
void IKFKSolver::NotifySubAnimChange()
{
	INode *targetNode=GetTargetNode();
	INode *mainNode=GetMainNode();
	INode *secNode=GetSecNode();
	INode *lastNode=GetLastNode();
	INode *goalNode=GetGoalNode();
	INode *goalParentNode;
	int nGoalParent=GoalParentCount();
	if (targetNode)
		targetNode->GetTMController()->NotifyDependents(FOREVER,PART_TM,REFMSG_SUBANIM_STRUCTURE_CHANGED);
	if (nGoalParent)
	{
		for (int i=0;i<nGoalParent;i++)
		{
			goalParentNode=GetGoalParentNode(i);
			if (goalParentNode)
				goalParentNode->GetTMController()->NotifyDependents(FOREVER,PART_TM,REFMSG_SUBANIM_STRUCTURE_CHANGED);	
		}
	}
	if (goalNode)
		goalNode->GetTMController()->NotifyDependents(FOREVER,PART_TM,REFMSG_SUBANIM_STRUCTURE_CHANGED);
	if (mainNode && secNode && lastNode)
	{
		mainNode->GetTMController()->NotifyDependents(FOREVER,PART_TM,REFMSG_SUBANIM_STRUCTURE_CHANGED);
		secNode->GetTMController()->NotifyDependents(FOREVER,PART_TM,REFMSG_SUBANIM_STRUCTURE_CHANGED);
		lastNode->GetTMController()->NotifyDependents(FOREVER,PART_TM,REFMSG_SUBANIM_STRUCTURE_CHANGED);
	}
}
//NotifyKeySelChange
//this is used to notify a key selection change message to the dependents of the controls that are part of the solver.
//this is necesary when using the show all keys mechanism.
void IKFKSolver::NotifyKeySelChange()
{
	if (notNotifyingKeySel) //used to avoid dependency loops
	{
		notNotifyingKeySel=FALSE;
		INode *targetNode=GetTargetNode();
		INode *mainNode=GetMainNode();
		INode *secNode=GetSecNode();
		INode *lastNode=GetLastNode();
		INode *goalNode=GetGoalNode();
		INode *goalParentNode;
		int nGoalParent=GoalParentCount();
		if (targetNode)
			targetNode->GetTMController()->NotifyDependents(FOREVER,PART_TM,REFMSG_KEY_SELECTION_CHANGED);
		if (nGoalParent)
		{
			for (int i=0;i<nGoalParent;i++)
			{
				goalParentNode=GetGoalParentNode(i);
				if (goalParentNode)
					goalParentNode->GetTMController()->NotifyDependents(FOREVER,PART_TM,REFMSG_KEY_SELECTION_CHANGED);	
			}
		}
		if (goalNode)
			goalNode->GetTMController()->NotifyDependents(FOREVER,PART_TM,REFMSG_KEY_SELECTION_CHANGED);
		if (mainNode && secNode && lastNode)
		{
			mainNode->GetTMController()->NotifyDependents(FOREVER,PART_TM,REFMSG_KEY_SELECTION_CHANGED);
			secNode->GetTMController()->NotifyDependents(FOREVER,PART_TM,REFMSG_KEY_SELECTION_CHANGED);
			lastNode->GetTMController()->NotifyDependents(FOREVER,PART_TM,REFMSG_KEY_SELECTION_CHANGED);
		}
		notNotifyingKeySel=TRUE;
	}
}

//GetGoalParentTM
//it is used to retrieve the tm of the parent of the ik goal, if  a goal parent is assigned
//it calculates the tm by using the local ik tm controllers
Matrix3 IKFKSolver::GetGoalParentTM(TimeValue t)
{
	int nGoalParent=GoalParentCount();
	
	if (nGoalParent)
	{
		INode *goalParentNode=GetGoalParentNode(0);
		Matrix3 goalParentTM;
		INode *theNode;
		Control *theCtrl;

		goalParentTM=GetParentTM(t,GetGoalParentNode(0),FOREVER);
		for (int i=0;i<nGoalParent;i++)
		{
			theNode=GetGoalParentNode(i);
			if (theNode)
			{
				theCtrl=GetGoalParentNode(i)->GetTMController();
				if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
					theCtrl=((IKCtrl*)theCtrl)->IK;
				if (theCtrl)
					theCtrl->GetValue(t,&goalParentTM,FOREVER,CTRL_RELATIVE);
			}			
		}
		return goalParentTM;
	}
	else
		return GetParentTM(t,GetGoalNode(),FOREVER);
}
//GetTargetParentTM
//it is used to retrieve the tm of the parent of the target node, if  a goal parent is part of the hierarchy
//it calculates the tm by using the local ik tm controllers
Matrix3 IKFKSolver::GetTargetParentTM(TimeValue t, INode *theNode, Interval &valid)
{
	if (theNode)
	{
		INode *parentNode=theNode->GetParentNode();
		if (parentNode && !parentNode->IsRootNode())
		{
			Matrix3 theTM;
			Control *theCtrl;
			theTM=GetTargetParentTM(t,parentNode,valid);
			theCtrl=parentNode->GetTMController();
			if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
				theCtrl=((IKCtrl*)theCtrl)->IK;
			theCtrl->GetValue(t,&theTM,valid,CTRL_RELATIVE);	
			return theTM;
		}
		else
			return Matrix3(1);
	}
	else
		return Matrix3(1);
}
//GetParentTM
//this function is used instead of the INode->GetParentTM because this function can avoid
//dependency loops
Matrix3 IKFKSolver::GetParentTM(TimeValue t,INode *theNode,Interval &valid)
{
	if (theNode)
	{
		INode *parentNode=theNode->GetParentNode();
		if (parentNode && !parentNode->IsRootNode())
		{
			Matrix3 theTM;
			theTM=GetParentTM(t,parentNode,valid);
			parentNode->GetTMController()->GetValue(t,&theTM,valid,CTRL_RELATIVE);	
			return theTM;
		}
		else
			return Matrix3(1);
	}
	else
		return Matrix3(1);
}

/*
FUNCTION IKFKSolver::MakeLookAtMatrix
Job:
	creates a matrix that have a position in the origin value, has its x-axis looking
	to the target value, and the y-axis looking to the lookup value
Parameters:
	Point3 origin:
		the position of the matrix
	Point3 target:
		the target of the lookat
	Point3 lookup:
		lookat uptarget position
*/
Matrix3 IKFKSolver::MakeLookAtMatrix(Point3 origin, Point3 target,Point3 lookup)
{
	Point3 vectorA=Normalize(origin-lookup);
	Point3 vectorB=Normalize(target-origin);
	Matrix3 tm;
	tm.SetRow(0,vectorB);
	tm.SetRow(2,Normalize(CrossProd(vectorB,vectorA)));
	tm.SetRow(1,Normalize(CrossProd(tm.GetRow(2),tm.GetRow(0))));
	tm.SetRow(3,origin);
	return tm;
}
/*
FUNCTION IKFKSolver::MakeLookAtMatrix2
Job:
	the same as the previous function but it takes 2 vectors as input and the matrix position
Parameters:
	Point3 lookVector:
		where the x axis is aiming
	Point3 upVector:
		the up vector (potential y)
	Point3 position:
		position of the matrix
*/
Matrix3 IKFKSolver::MakeLookAtMatrix2(Point3 lookVector, Point3 upVector,Point3 position)
{
	lookVector=Normalize(lookVector);
	upVector=Normalize(upVector);
	Matrix3 tm;
	tm.SetRow(0,lookVector);
	tm.SetRow(2,Normalize(CrossProd(lookVector,upVector)));
	tm.SetRow(1,Normalize(CrossProd(tm.GetRow(2),tm.GetRow(0))));
	tm.SetRow(3,position);
	return tm;
}
/*
FUNCTION IKFKSolver::MakeLookAtQuat
Job:
	creates a Quaternion that is looking the target value, its building its made
	having the x-axis of the lookup as origin
Parameters:
	Point3 origin:
		the position of the lookat
	Point3 target:
		the target of the lookat
	Point3 lookup:
		lookat uptarget matrix
*/
Quat IKFKSolver::MakeLookAtQuat(Point3 origin, Point3 target,Matrix3 lookup)
{
	lookup.Orthogonalize();
	lookup.NoScale();
	Point3 vectorA=Point3(1.0f,0.0f,0.0f);
	Point3 vectorB=Normalize(Inverse(lookup).VectorTransform(target-origin));
	return Quat(AngAxis(Normalize(CrossProd(vectorA,vectorB)),-acos(DotProd(vectorA,vectorB))))*Quat(lookup);
}
/*
FUNCTION IKFKSolver::RotVector
Job:
	Finds a Quat that rotates a vector to another.
Parameters:
	Point3 v1:
		vector to rotate
	Point3 v2:
		target vector
*/
Quat IKFKSolver::RotVector(Point3 v1, Point3 v2)
{
	v1=Normalize(v1);
	v2=Normalize(v2);
	return Quat(AngAxis(Normalize(CrossProd(v1,v2)),-acos(DotProd(v1,v2))));
}

float IKFKSolver::GetCloserAngle(float prevAngle, float newAngle)
{
	float modPrevAngle, theDif;
	modPrevAngle=TWOPI*modf(prevAngle/TWOPI,new FLOAT());
	if (modPrevAngle<0.0f) modPrevAngle+=TWOPI;
	theDif=newAngle-modPrevAngle;
	if (theDif>PI)
		theDif=theDif-TWOPI;
	else if (theDif<-PI)
		theDif=TWOPI+theDif;
	return (prevAngle+theDif);
}
//GetCtrlPRS
//this function retrievs the position, rotation and scale values of the defined Matrix Controller
//by using its position, rotation and scale controllers instead of the TM controller itself
void IKFKSolver::GetCtrlPRS(TimeValue t,Control *ctrl,Point3 &pos,Quat &rot,Point3 &scale, Interval &valid)
{
	if (ctrl)
	{
		Control *posCtrl=ctrl->GetPositionController();
		Control *rotCtrl=ctrl->GetRotationController();
		Control *scaleCtrl=ctrl->GetScaleController();
		if (posCtrl && rotCtrl && scaleCtrl)
		{
			ScaleValue theScale;
			posCtrl->GetValue(t,&pos,valid,CTRL_ABSOLUTE);
			rotCtrl->GetValue(t,&rot,valid,CTRL_ABSOLUTE);
			scaleCtrl->GetValue(t,&theScale,valid,CTRL_ABSOLUTE);
			scale=theScale.s;
		}
		else
		{
			Matrix3 theTM=Matrix3(1);
			ctrl->GetValue(t,&theTM,valid,CTRL_RELATIVE);
			pos=theTM.GetTrans();
			rot=MatrixToQuat(theTM);
			scale=GetTMScale(theTM);
		}
	}
	else
	{
		pos=Point3(0.0f,0.0f,0.0f);
		rot=Quat(0.0f,0.0f,0.0f,1.0f);
		scale=Point3(1.0f,1.0f,1.0f);
	}
}
Point3 IKFKSolver::GetTMScale(Matrix3 theTM)
{
	Point3 theS,theP;
	Quat theQ;
	DecomposeMatrix(theTM,theP,theQ,theS);
	return  theS;
	//return Point3(theTM.GetRow(0).FLength(),theTM.GetRow(1).FLength(),theTM.GetRow(2).FLength());
}


//This function is used because the are problems when a quaternion is created using a matrix
//with is default creation function Quat(const Matrix3 &mat) if the matrix is scaled or skewed
Quat IKFKSolver::MatrixToQuat(Matrix3 theTM)
{
	theTM.Orthogonalize();
	theTM.NoScale();
	return Quat(theTM);
}
void IKFKSolver::CheckHierarchy()
{
	if (GoalParentCount())
	{
		if (!isValidGoalParent(GetGoalParentNode(0)))
		{
			BOOL holdSuspended=theHold.IsSuspended();
			if (holdSuspended) theHold.Resume();
			SetGoalParentNode(NULL);
			if (holdSuspended) theHold.Suspend();
		}
	}
}
//canupdate is used to shut down the messaging system of the solver when it's needed
BOOL IKFKSolver::CanUpdate()
{
	return canUpdate;
}
void IKFKSolver::SetCanUpdate(BOOL canUpdate)
{
	this->canUpdate=canUpdate;
}

//isValidTarget
//the target cannot be a part of other solver (or itself)
BOOL IKFKSolver::isValidTarget(INode* theNode)
{
	return (theNode && theNode->GetTMController()->ClassID()!=IKCtrl_CLASS_ID && theNode->GetTMController()->ClassID()!=FKCtrl_CLASS_ID);
		/*
	targetNode->BeginDependencyTest();
	solver->NotifyDependents(FOREVER,0,REFMSG_TEST_DEPENDENCY);
	return (!targetNode->EndDependencyTest() && (theAnimPivot->pblock->GetINode(pb_target)!=targetNode));
	*/
}

//isValidGoalParent
//the goal parent has to be a node that is above in the hierarchy of the Ik goal
BOOL IKFKSolver::isValidGoalParent(INode* theNode)
{
	BOOL notFound=TRUE;
	INode *aux;
	aux=GetGoalNode();
	if (aux)
	{
		aux=aux->GetParentNode();
		while (!aux->IsRootNode())
		{
			BOOL isAuxGPN=FALSE;
			for (int i=0;i<GoalParentCount();i++)
				if (aux==GetGoalParentNode(i))
					isAuxGPN=TRUE;
			if (!isAuxGPN && (aux->GetTMController()->ClassID()==IKCtrl_CLASS_ID || aux->GetTMController()->ClassID()==FKCtrl_CLASS_ID))
				break;
			if (aux==theNode)
			{
				notFound=FALSE;
				break;
			}
			aux=aux->GetParentNode();
		}
	}
	return !notFound;
}
//display lines showing the current transformation of the local fk-ik values
BOOL IKFKSolver::DisplayIKFK(TimeValue t,INode* inode,ViewExp* vpt,int flags)
{
	BOOL canDisplay=FALSE;
	if (pblock)
		pblock->GetValue(pb_displayIKFK,0,canDisplay,FOREVER);
	
	if (canDisplay && vpt)
	{
		INode *goalNode, *mainNode, *secNode, *lastNode;
		goalNode=GetGoalNode();
		mainNode=GetMainNode();
		secNode=GetSecNode();
		lastNode=GetLastNode();

		if (goalNode && mainNode && secNode && lastNode)
		{
			Matrix3 goaltm,parenttm,maintm,sectm,lasttm;
			Control *theCtrl;

			goaltm=GetGoalParentTM(t);
			theCtrl=goalNode->GetTMController();
			if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
				theCtrl=((IKCtrl*)theCtrl)->IK;
			theCtrl->GetValue(t,&goaltm,FOREVER,CTRL_RELATIVE);

			parenttm=GetParentTM(t,GetMainNode(),FOREVER);


			maintm=parenttm;
			theCtrl=mainNode->GetTMController();
			if (theCtrl->ClassID()==FKCtrl_CLASS_ID)
				theCtrl=((FKCtrl*)theCtrl)->FK;
			if (theCtrl)
				theCtrl->GetValue(t,&maintm,FOREVER,CTRL_RELATIVE);

			sectm=maintm;
			theCtrl=secNode->GetTMController();
			if (theCtrl->ClassID()==FKCtrl_CLASS_ID)
				theCtrl=((FKCtrl*)theCtrl)->FK;
			if (theCtrl)
				theCtrl->GetValue(t,&sectm,FOREVER,CTRL_RELATIVE);

			lasttm=sectm;
			theCtrl=lastNode->GetTMController();
			if (theCtrl->ClassID()==FKCtrl_CLASS_ID)
				theCtrl=((FKCtrl*)theCtrl)->FK;
			if (theCtrl)
				theCtrl->GetValue(t,&lasttm,FOREVER,CTRL_RELATIVE);


			GraphicsWindow *gw = vpt->getGW();
			gw->setTransform(Matrix3(1));
			gw->setColor(LINE_COLOR, Point3(1.0f,0.0f,0.0f));
			Point3 points[4];
			points[0]=maintm.GetTrans();
			points[1]=sectm.GetTrans();
			points[2]=lasttm.GetTrans();
			gw->polyline( 3, points, NULL, NULL, FALSE, NULL );
			gw->setColor(LINE_COLOR, Point3(0.0f,0.0f,1.0f));
			points[0]=maintm.GetTrans();
			points[1]=goaltm.GetTrans();
			gw->polyline( 2, points, NULL, NULL, FALSE, NULL );
			

			float theLength=(sectm.GetTrans()-maintm.GetTrans()).FLength()*0.5f;
			gw->setColor(LINE_COLOR, Point3(0.0f,1.0f,0.0f));
			Point3 pref=Point3(theLength,0.0f,0.0f);
			Matrix3 prefTM;
			float prefRot[3];


			//Draw Main pref rotation
			theCtrl=GetMainNode()->GetTMController();
			prefRot[0]=0.0f;
			prefRot[1]=0.0f;
			prefRot[2]=0.0f;
			if (theCtrl->ClassID()==FKCtrl_CLASS_ID && ((FKCtrl*)theCtrl)->prefPblock)
			{
				((FKCtrl*)theCtrl)->prefPblock->GetValue(0,t,prefRot[0],FOREVER);
				((FKCtrl*)theCtrl)->prefPblock->GetValue(1,t,prefRot[1],FOREVER);
				((FKCtrl*)theCtrl)->prefPblock->GetValue(2,t,prefRot[2],FOREVER);
			}
			EulerToMatrix(prefRot,prefTM,EULERTYPE_XYZ);
			prefTM*=parenttm;
			prefTM.SetTrans(maintm.GetTrans());
			pref=pref*prefTM;
			points[0]=maintm.GetTrans();
			points[1]=pref;
			gw->polyline( 2, points, NULL, NULL, FALSE, NULL );

			pref=Point3(theLength,0.0f,0.0f);

			//Draw Sec pref rotation
			theCtrl=GetSecNode()->GetTMController();
			prefRot[0]=0.0f;
			prefRot[1]=0.0f;
			prefRot[2]=0.0f;
			if (theCtrl->ClassID()==FKCtrl_CLASS_ID && ((FKCtrl*)theCtrl)->prefPblock)
			{
				((FKCtrl*)theCtrl)->prefPblock->GetValue(0,t,prefRot[0],FOREVER);
				((FKCtrl*)theCtrl)->prefPblock->GetValue(1,t,prefRot[1],FOREVER);
				((FKCtrl*)theCtrl)->prefPblock->GetValue(2,t,prefRot[2],FOREVER);
			}
			EulerToMatrix(prefRot,prefTM,EULERTYPE_XYZ);
			prefTM*=maintm;
			prefTM.SetTrans(sectm.GetTrans());
			pref=pref*prefTM;
			points[0]=sectm.GetTrans();
			points[1]=pref;
			gw->polyline( 2, points, NULL, NULL, FALSE, NULL );
			return TRUE;
		}
		return FALSE;
	}
	return FALSE;
}

//FindCtrl
//checks if the controller is already in the ctrl array
BOOL IKFKSolver::FindCtrl (Animatable *ctrl, Tab<Animatable*> &theCtrls)
{
	for (int i=0;i<theCtrls.Count();i++)
		if (theCtrls[i]==ctrl)
			return TRUE;
	return FALSE;
}
//AppendKeyableCtrls
//used to recursively get the keyable controllers from the controller sub anim tree
void IKFKSolver::AppendKeyableCtrls (Animatable *ctrl, Tab<Animatable*> &theCtrls)
{
	if (ctrl)
	{
		//check if it's keyframable and if it's not already appended
		if (ctrl->NumKeys()!=NOT_KEYFRAMEABLE && !FindCtrl(ctrl,theCtrls))
		{
			//now check if it's not reference of a selected node hierarchy (to avoid multiple keyframe mapping)
			INodeTab dependentNodes;
			bool isSelected=FALSE;
			FindNodesProc dep(&dependentNodes);
			((ReferenceTarget*)ctrl)->DoEnumDependents(&dep);

			for (int i=0;i<dependentNodes.Count();i++)
				if (dependentNodes[i]->Selected())
					isSelected=TRUE;

			if (!isSelected)
				theCtrls.Append(1,&ctrl);
		}
		for (int i=0;i<ctrl->NumSubs();i++)
			AppendKeyableCtrls(ctrl->SubAnim(i),theCtrls);
	}
}
//RefreshKeys
// updates the keyable controllers in the allCtrls array which is used in the show all keys mechanism
//the allkeys mechanism is activated in only one node to avoid multiple keys transformation in the track bar
void IKFKSolver::RefreshKeys()
{
	BOOL showAllKeys=FALSE;
	if (pblock)
		pblock->GetValue(pb_showAllKeys,0,showAllKeys,FOREVER);	
	INode* theNode;
	Control *theCtrl;

	//each controller have a flag called allKeysActive, they only make keyframe operations if that flag is on,
	//that flag is used to avoid multiple keyframe operations in the same solver (to avoid multiple key mapping)
	//only one node will have that flag on if several of them are selected
	if (showAllKeys)
	{
		allCtrls.Init();

		BOOL alreadyActivated=FALSE;
		theNode=GetMainNode();
		if (theNode)
		{
			theCtrl=theNode->GetTMController();
			if (theCtrl->ClassID()==FKCtrl_CLASS_ID)
			{
				if (theNode->Selected())
				{
					if (!alreadyActivated)
					{
						((FKCtrl*)theCtrl)->allKeysActive=TRUE;
						alreadyActivated=TRUE;
					}
					else
						((FKCtrl*)theCtrl)->allKeysActive=FALSE;
				}
				else
				{
					AppendKeyableCtrls((Animatable*)((FKCtrl*)theCtrl)->FK,allCtrls);
					AppendKeyableCtrls((Animatable*)((FKCtrl*)theCtrl)->prefPblock,allCtrls);
					((FKCtrl*)theCtrl)->allKeysActive=FALSE;
				}
			}
		}
		theNode=GetSecNode();
		if (theNode)
		{
			theCtrl=theNode->GetTMController();
			if (theCtrl->ClassID()==FKCtrl_CLASS_ID)
			{
				if (theNode->Selected())
				{
					if (!alreadyActivated)
					{
						((FKCtrl*)theCtrl)->allKeysActive=TRUE;
						alreadyActivated=TRUE;
					}
					else
						((FKCtrl*)theCtrl)->allKeysActive=FALSE;
				}
				else
				{
					AppendKeyableCtrls((Animatable*)((FKCtrl*)theCtrl)->FK,allCtrls);
					AppendKeyableCtrls((Animatable*)((FKCtrl*)theCtrl)->prefPblock,allCtrls);
					((FKCtrl*)theCtrl)->allKeysActive=FALSE;
				}
			}
		}
		theNode=GetLastNode();
		if (theNode)
		{
			theCtrl=theNode->GetTMController();
			if (theCtrl->ClassID()==FKCtrl_CLASS_ID)
			{
				if (theNode->Selected())
				{
					if (!alreadyActivated)
					{
						((FKCtrl*)theCtrl)->allKeysActive=TRUE;
						alreadyActivated=TRUE;
					}
					else
						((FKCtrl*)theCtrl)->allKeysActive=FALSE;
				}
				else
				{
					AppendKeyableCtrls((Animatable*)((FKCtrl*)theCtrl)->FK,allCtrls);
					AppendKeyableCtrls((Animatable*)((FKCtrl*)theCtrl)->prefPblock,allCtrls);
					((FKCtrl*)theCtrl)->allKeysActive=FALSE;
				}
			}
		}
		theNode=GetGoalNode();
		if (theNode)
		{
			theCtrl=theNode->GetTMController();
			if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
			{
				if (theNode->Selected())
				{
					if (!alreadyActivated)
					{
						((IKCtrl*)theCtrl)->allKeysActive=TRUE;
						alreadyActivated=TRUE;
					}
					else
						((IKCtrl*)theCtrl)->allKeysActive=FALSE;
				}
				else
				{
					AppendKeyableCtrls((Animatable*)((IKCtrl*)theCtrl)->IK,allCtrls);
					((IKCtrl*)theCtrl)->allKeysActive=FALSE;
				}
			}
		}
		theNode=GetTargetNode();
		if (theNode)
		{
			theCtrl=theNode->GetTMController();
			if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
			{
				if (theNode->Selected())
				{
					if (!alreadyActivated)
					{
						((IKCtrl*)theCtrl)->allKeysActive=TRUE;
						alreadyActivated=TRUE;
					}
					else
						((IKCtrl*)theCtrl)->allKeysActive=FALSE;
				}
				else
				{
					AppendKeyableCtrls((Animatable*)((IKCtrl*)theCtrl)->IK,allCtrls);
					((IKCtrl*)theCtrl)->allKeysActive=FALSE;
				}
			}
		}
		for (int i=0;i<GoalParentCount();i++)
		{
			theNode=GetGoalParentNode(i);
			if (theNode)
			{
				theCtrl=theNode->GetTMController();
				if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
				{
					if (theNode->Selected())
					{
						if (!alreadyActivated)
						{
							((IKCtrl*)theCtrl)->allKeysActive=TRUE;
							alreadyActivated=TRUE;
						}
						else
							((IKCtrl*)theCtrl)->allKeysActive=FALSE;
					}
					else
					{
						AppendKeyableCtrls((Animatable*)((IKCtrl*)theCtrl)->IK,allCtrls);
						((IKCtrl*)theCtrl)->allKeysActive=FALSE;
					}
				}
			}
		}
	}
	else //if the option show all keys is off all the ctrls will have the allKeysActive flag off
	{
		theNode=GetMainNode();
		if (theNode)
		{
			theCtrl=theNode->GetTMController();
			if (theCtrl->ClassID()==FKCtrl_CLASS_ID)
				((FKCtrl*)theCtrl)->allKeysActive=FALSE;
		}
		theNode=GetSecNode();
		if (theNode)
		{
			theCtrl=theNode->GetTMController();
			if (theCtrl->ClassID()==FKCtrl_CLASS_ID)
				((FKCtrl*)theCtrl)->allKeysActive=FALSE;
		}
		theNode=GetLastNode();
		if (theNode)
		{
			theCtrl=theNode->GetTMController();
			if (theCtrl->ClassID()==FKCtrl_CLASS_ID)
				((FKCtrl*)theCtrl)->allKeysActive=FALSE;
		}
		theNode=GetGoalNode();
		if (theNode)
		{
			theCtrl=theNode->GetTMController();
			if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
				((IKCtrl*)theCtrl)->allKeysActive=FALSE;
		}
		theNode=GetTargetNode();
		if (theNode)
		{
			theCtrl=theNode->GetTMController();
			if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
				((IKCtrl*)theCtrl)->allKeysActive=FALSE;
		}
		for (int i=0;i<GoalParentCount();i++)
		{
			theNode=GetGoalParentNode(i);
			if (theNode)
			{
				theCtrl=theNode->GetTMController();
				if (theCtrl->ClassID()==IKCtrl_CLASS_ID)
					((IKCtrl*)theCtrl)->allKeysActive=FALSE;
			}
		}
	}
}
//GetLocalKeyIndex
//used in keyframe functions, the solver will manipulate several controllers at the same time, so it has to make a translation system
//of the key index
int  IKFKSolver::GetLocalKeyIndex(int index, int &ctrlIndex)
{
	for (ctrlIndex=0;ctrlIndex<allCtrls.Count();ctrlIndex++)
	{
		int nKeys=allCtrls[ctrlIndex]->NumKeys();
		if (index>=nKeys)
			index-=nKeys;
		else
			return index;
	}
	return -1;
}

int  IKFKSolver::NumKeys ()
{
	int keyCount=0;
	for (int i=0;i<allCtrls.Count();i++)
		keyCount+=allCtrls[i]->NumKeys();
	return keyCount;
}
TimeValue  IKFKSolver::GetKeyTime (int index)
{
	int localIndex;
	int ctrlIndex;
	localIndex=GetLocalKeyIndex(index,ctrlIndex);
	if (localIndex>=0)
		return allCtrls[ctrlIndex]->GetKeyTime(localIndex);
	else
		return 0;
}
int  IKFKSolver::GetKeyIndex (TimeValue t)
{
	for (int i=0;i<allCtrls.Count();i++)
	{
		int keyIndex=allCtrls[i]->GetKeyIndex(t);
		if (keyIndex>=0)
			return keyIndex;
	}
	return -1;
}
BOOL  IKFKSolver::GetNextKeyTime (TimeValue t, DWORD flags, TimeValue &nt) 
{
	nt=t;
	TimeValue currentT=t;
	BOOL flag=FALSE;

	if (flags & NEXTKEY_RIGHT)
	{
		for (int i=0;i<allCtrls.Count();i++)
			if (allCtrls[i]->GetNextKeyTime (t,flags,currentT) )
				if (!flag)
				{
					nt=currentT;
					flag=TRUE;
				}
				else
					if (currentT>t)
						if (nt>t)
						{
							if (currentT<nt)
								nt=currentT;
						}
						else
							nt=currentT;
					else
						if (nt<t)
							if (currentT<nt)
								nt=currentT;
	}
	else if (flags & NEXTKEY_LEFT)
	{
		for (int i=0;i<allCtrls.Count();i++)
			if (allCtrls[i]->GetNextKeyTime (t,flags,currentT) )
				if (!flag)
				{
					nt=currentT;
					flag=TRUE;
				}
				else
					if (currentT<t)
						if (nt<t)
						{
							if (currentT>nt)
								nt=currentT;
						}
						else
							nt=currentT;
					else
						if (nt>t)
							if (currentT>nt)
								nt=currentT;
	}
	return flag;
}
void  IKFKSolver::CopyKeysFromTime (TimeValue src, TimeValue dst, DWORD flags)
{
	for (int i=0;i<allCtrls.Count();i++)
		allCtrls[i]->CopyKeysFromTime(src,dst,flags);
}
void  IKFKSolver::DeleteKeyAtTime (TimeValue t)
{
	for (int i=0;i<allCtrls.Count();i++)
		allCtrls[i]->DeleteKeyAtTime(t);
}
BOOL  IKFKSolver::IsKeyAtTime (TimeValue t, DWORD flags)
{
	BOOL flag=0;
	for (int i=0;i<allCtrls.Count();i++)
		if (allCtrls[i]->IsKeyAtTime(t,flags))
			flag=TRUE;
	return flag;
}
int  IKFKSolver::GetKeyTimes (Tab< TimeValue > &times, Interval range, DWORD flags)
{
	return 0;
}
int  IKFKSolver::GetKeySelState (BitArray &sel, Interval range, DWORD flags)
{
	return 0;
}
void  IKFKSolver::MapKeys (TimeMap *map, DWORD flags)
{
	for (int i=0;i<allCtrls.Count();i++)
		allCtrls[i]->MapKeys(map,flags);
}
void  IKFKSolver::DeleteKeys (DWORD flags)
{
	for (int i=0;i<allCtrls.Count();i++)
		allCtrls[i]->DeleteKeys(flags);
}
void  IKFKSolver::DeleteKeyByIndex (int index)
{
	int localIndex;
	int ctrlIndex;
	localIndex=GetLocalKeyIndex(index,ctrlIndex);
	if (localIndex>=0)
		allCtrls[ctrlIndex]->DeleteKeyByIndex(localIndex);
}
void  IKFKSolver::SelectKeys (TrackHitTab &sel, DWORD flags)
{
	for (int i=0;i<allCtrls.Count();i++)
		allCtrls[i]->SelectKeys(sel,flags);
}
void  IKFKSolver::SelectKeyByIndex (int i, BOOL sel)
{
	int localIndex;
	int ctrlIndex;
	localIndex=GetLocalKeyIndex(i,ctrlIndex);
	if (localIndex>=0)
		allCtrls[ctrlIndex]->SelectKeyByIndex(localIndex,sel);
}
BOOL  IKFKSolver::IsKeySelected (int i)
{
	int localIndex;
	int ctrlIndex;
	localIndex=GetLocalKeyIndex(i,ctrlIndex);
	if (localIndex>=0)
		return allCtrls[ctrlIndex]->IsKeySelected(localIndex);
	else
		return FALSE;
}
void  IKFKSolver::FlagKey (TrackHitRecord hit)
{
	for (int i=0;i<allCtrls.Count();i++)
		allCtrls[i]->FlagKey(hit);
}
int  IKFKSolver::GetFlagKeyIndex ()
{
	for (int i=0;i<allCtrls.Count();i++)
	{
		int index=allCtrls[i]->GetFlagKeyIndex();
		if (index>=0)
			return index;
	}
	return -1;
}
int  IKFKSolver::NumSelKeys ()
{
	int keyCount=0;
	for (int i=0;i<allCtrls.Count();i++)
		keyCount+=allCtrls[i]->NumSelKeys();
	return keyCount;
}
void  IKFKSolver::CloneSelectedKeys (BOOL offset)
{
	for (int i=0;i<allCtrls.Count();i++)
		allCtrls[i]->CloneSelectedKeys(offset);
}
void  IKFKSolver::AddNewKey (TimeValue t, DWORD flags)
{
	for (int i=0;i<allCtrls.Count();i++)
		allCtrls[i]->AddNewKey(t,flags);
}